from abc import ABC, abstractmethod
import os
import sys
import time
import glob
import h5py
import numcodecs
import argparse
import datetime
import subprocess
import numpy as np
from sklearn.utils import shuffle
from pprint import pprint
# silence tensorflow warnings
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
import warnings

with warnings.catch_warnings():
    warnings.filterwarnings("ignore")
    import tensorflow as tf
    from tensorflow_addons.optimizers import AdamW
    from tensorflow.keras.callbacks import Callback
    from tensorflow.keras import optimizers
    from tensorflow.keras import backend as K
    from tensorflow.keras.models import load_model, Model
    from tensorflow.keras.utils import Sequence
    from tensorflow.keras.layers import Input

from .metrics import Metrics
from .overlap import OverlapSeqHelper
from .helpers import get_contiguous_ranges


class ConfusionMatrixTrain(Callback):
    def __init__(
        self,
        save_model_path,
        train_generator,
        val_generator,
        large_eval_folder,
        patience,
        calc_H=False,
        report_to_nni=False,
        check_every_nth_batch=1_000_000,
        save_every_check=False,
    ):
        self.save_model_path = save_model_path
        self.save_dir = os.path.dirname(save_model_path)
        self.save_every_check = save_every_check
        self.train_generator = train_generator
        self.val_generator = val_generator
        self.large_eval_folder = large_eval_folder
        self.patience = patience
        self.calc_H = calc_H
        self.report_to_nni = report_to_nni
        self.best_val_genic_f1 = 0.0
        self.checks_without_improvement = 0
        self.check_every_nth_batch = (
            check_every_nth_batch  # high default for ~ 1 / epoch
        )
        self.epoch = 0
        print(self.save_model_path, "SAVE MODEL PATH")

    def on_epoch_begin(self, epoch, logs=None):
        self.epoch_start = time.time()

    def on_epoch_end(self, epoch, logs=None):
        print(f"training took {(time.time() - self.epoch_start) / 60:.2f}m")
        self.check_in()
        self.epoch += 1

    def on_train_batch_end(self, batch, logs=None):
        if not (batch + 1) % self.check_every_nth_batch:
            print(f"\nvalidation and checkpoint at batch {batch}")
            self.check_in(batch)

    def freeze_layers(self, model):
        # thank you https://github.com/keras-team/keras/issues/13279#issuecomment-527705263
        for i in model.layers:
            i.trainable = False
            if isinstance(i, Model):
                self.freeze_layers(i)
        return model

    def check_in(self, batch=None):
        _, _, val_genic_f1 = HelixerModel.run_metrics(
            self.val_generator, self.model, calc_H=self.calc_H
        )

        if val_genic_f1 > self.best_val_genic_f1:
            self.best_val_genic_f1 = val_genic_f1
            self.freeze_layers(self.model)
            self.model.save(self.save_model_path, save_format="h5")
            print(
                "saved new best model with genic f1 of {} at {}".format(
                    self.best_val_genic_f1, self.save_model_path
                )
            )
            self.checks_without_improvement = 0
        else:
            self.checks_without_improvement += 1
            if self.checks_without_improvement >= self.patience:
                self.model.stop_training = True
        if batch is None:
            b_str = "epoch_end"
        else:
            b_str = f"b{batch:06}"
        if self.save_every_check:
            path = os.path.join(self.save_dir, f"model_e{self.epoch}_{b_str}.h5")
            self.model.save(path, save_format="h5")
            print(f"saved model at {path}")

    def on_train_end(self, logs=None):
        if os.path.isdir(self.large_eval_folder):
            # load best model
            best_model = load_model(self.save_model_path)
            # double check that we loaded the correct model, can be remove if confirmed this works
            print("\nValidation set again:")
            _, _, val_genic_f1 = HelixerModel.run_metrics(
                self.val_generator, best_model, print_to_stdout=True, calc_H=self.calc_H
            )
            assert val_genic_f1 == self.best_val_genic_f1

            training_species = self.train_generator.h5_file.attrs["genomes"]
            median_f1 = HelixerModel.run_large_eval(
                self.large_eval_folder, best_model, self.val_generator, training_species
            )


class PreshuffleCallback(Callback):
    def __init__(self, train_generator):
        self.train_generator = train_generator

    def on_epoch_begin(self, epoch, logs=None):
        if self.train_generator.shuffle:
            self.train_generator.shuffle_data()


class HelixerSequence(Sequence):
    def __init__(self, model, h5_files, mode, batch_size, shuffle):
        assert mode in ["train", "val", "test"]
        self.model = model
        self.h5_files = h5_files
        self.mode = mode
        self.shuffle = shuffle
        self.batch_size = batch_size
        self._cp_into_namespace(
            [
                "float_precision",
                "class_weights",
                "transition_weights",
                "input_coverage",
                "coverage_count",
                "coverage_norm",
                "overlap",
                "overlap_offset",
                "core_length",
                "stretch_transition_weights",
                "coverage_weights",
                "coverage_offset",
                "no_utrs",
                "predict_phase",
                "load_predictions",
                "only_predictions",
                "debug",
            ]
        )

        if self.mode == "test":
            assert (
                len(self.h5_files) == 1
            ), "predictions and eval should be applied to individual files only"

        # set chunk size and overlap parameters
        # this is first done here, because its pulled dynamically from h5 files
        chunk_sizes = [h5["data/X"].shape[1] for h5 in self.h5_files]
        for cs in chunk_sizes[1:]:
            assert (
                cs == chunk_sizes[0]
            ), f"Not all subsequence lengths match in h5_files: {chunk_sizes}"
        self.chunk_size = chunk_sizes[0]
        # once we have the chunk_size, we can find defaults for overlapping
        if self.overlap_offset is None:
            self.overlap_offset = self.chunk_size // 2
        if self.core_length is None:
            self.core_length = int(self.chunk_size * 3 / 4)

        self.data_list_names = ["data/X"]
        if not self.only_predictions:
            self.data_list_names += ["data/y", "data/sample_weights"]
            if self.load_predictions:
                self.data_list_names.append("data/predictions")
            if self.predict_phase:
                self.data_list_names.append("data/phases")
            if self.mode == "train":
                if self.transition_weights is not None:
                    self.data_list_names.append("data/transitions")
                if self.coverage_weights:
                    self.data_list_names.append("scores/by_bp")

        if self.overlap:
            assert (
                self.mode == "test"
            ), "overlapping currently only works for test (predictions & eval)"
            # can take [0] below bc we've asserted that test means len(self.h5_files) == 1 above
            contiguous_ranges = get_contiguous_ranges(self.h5_files[0])
            self.ol_helper = OverlapSeqHelper(
                contiguous_ranges=contiguous_ranges,
                chunk_size=self.chunk_size,
                max_batch_size=self.batch_size,
                overlap_offset=self.overlap_offset,
                core_length=self.core_length,
            )

        if self.input_coverage:
            self.data_list_names += [
                "evaluation/rnaseq_coverage",
                "evaluation/rnaseq_spliced_coverage",
            ]

        self.data_lists = [[] for _ in range(len(self.data_list_names))]
        self.data_dtypes = [
            self.h5_files[0][name].dtype for name in self.data_list_names
        ]

        self.compressor = numcodecs.blosc.Blosc(
            cname="blosclz", clevel=4, shuffle=2
        )  # use BITSHUFFLE

        for h5_file in self.h5_files:
            self._load_one_h5(h5_file)

        for name, data_list in zip(self.data_list_names, self.data_lists):
            comp_data_size = sum([sys.getsizeof(e) for e in data_list])

        self.n_seqs = len(self.data_lists[0])

        if self.mode == "test":
            if self.class_weights is not None:
                print(
                    f'ignoring the class_weights of {self.class_weights} in mode "test"'
                )
                self.class_weights = None
            if self.transition_weights is not None:
                print(
                    f'ignoring the transition_weights of {self.transition_weights} in mode "test"'
                )
                self.transition_weights = None

    def _load_one_h5(self, h5_file):
        x_dset = h5_file["data/X"]
        if not self.only_predictions:
            y_dset = h5_file["data/y"]
            print(f"y shape: {y_dset.shape}")

        if self.debug:
            # so that total sequences between all files add to ~1000
            n_seqs = max(1000 // len(self.h5_files), 1)
        else:
            n_seqs = x_dset.shape[0]

        if self.mode == "train" or self.mode == "val":
            mask = np.logical_and(
                h5_file["data/is_annotated"], h5_file["data/err_samples"]
            )
            n_masked = x_dset.shape[0] - np.sum(mask)
            print(
                f"\nmasking {n_masked} completely un-annotated or completely erroneous sequences"
            )

        else:
            mask = np.ones(h5_file["data/X"].shape[0], dtype=bool)
            n_masked = 0

        # load at most 2000 uncompressed samples at a time in memory
        max_at_once = min(2000, n_seqs)
        for name, data_list in zip(self.data_list_names, self.data_lists):
            start_time_dset = time.time()
            for offset in range(0, n_seqs, max_at_once):
                step_mask = mask[offset : offset + max_at_once]
                if name == "data/predictions":
                    data_slice = h5_file[name][0, offset : offset + max_at_once][
                        step_mask
                    ]  # only use one prediction for now
                else:
                    data_slice = h5_file[name][offset : offset + max_at_once][step_mask]
                if self.no_utrs and name == "data/y":
                    HelixerSequence._zero_out_utrs(data_slice)
                data_list.extend([self.compressor.encode(e) for e in data_slice])

    @staticmethod
    def _zero_out_utrs(y):
        # merge UTR and IG labels and zero out the UTR column
        # still keep 4 columns for simplicity of downstream code and (maybe) more transfer learning potential
        y[..., 0] = np.logical_or(y[..., 0], y[..., 1])
        y[..., 1] = 0

    def shuffle_data(self):
        start_time = time.time()
        self.data_lists = shuffle(*self.data_lists)

    def _cp_into_namespace(self, names):
        """Moves class properties from self.model into this class for brevity"""
        for name in names:
            self.__dict__[name] = self.model.__dict__[name]

    def _get_batch_data(self, batch_idx):
        batch = []
        # batch must have one thing for everything unpacked by __getitem__ (and in order)
        for name in [
            "data/X",
            "data/y",
            "data/sample_weights",
            "data/transitions",
            "data/phases",
            "data/predictions",
            "scores/by_bp",
        ]:
            if name not in self.data_list_names:
                batch.append(None)
            else:
                decoded_list = self.get_batch_of_one_dataset(name, batch_idx)

                # append coverage to X directly, might be clearer elsewhere once working, but this needs little code...
                if name == "data/X" and self.input_coverage:
                    decode_coverage = self.get_batch_of_one_dataset(
                        "evaluation/rnaseq_coverage", batch_idx
                    )
                    decode_coverage = [
                        self._cov_norm(x.reshape(-1, self.coverage_count)).astype(
                            np.float16
                        )
                        for x in decode_coverage
                    ]
                    decode_spliced = self.get_batch_of_one_dataset(
                        "evaluation/rnaseq_spliced_coverage", batch_idx
                    )
                    decode_spliced = [
                        self._cov_norm(x.reshape(-1, self.coverage_count)).astype(
                            np.float16
                        )
                        for x in decode_spliced
                    ]
                    decoded_list = [
                        np.concatenate((x, y, z), axis=1)
                        for x, y, z in zip(
                            decoded_list, decode_coverage, decode_spliced
                        )
                    ]

                decoded = np.stack(decoded_list, axis=0)
                if self.overlap and name == "data/X":
                    decoded = self.ol_helper.make_input(batch_idx, decoded)

                batch.append(decoded)

        return tuple(batch)

    def get_batch_of_one_dataset(self, name, batch_idx):
        """returns single batch (the Nth where N=batch_idx) from dataset '{name}'"""
        # setup indices based on overlapping or not
        if self.overlap:
            h5_indices = self.ol_helper.h5_indices_of_batch(batch_idx)
        else:
            end = min(self.n_seqs, (batch_idx + 1) * self.batch_size)
            h5_indices = np.arange(batch_idx * self.batch_size, end)

        return self._decode_one(name, h5_indices)

    def _decode_one(self, name, h5_indices):
        """decode batch delineated by h5_indices from compressed data originally from dataset {name}"""
        i = self.data_list_names.index(name)
        dtype = self.data_dtypes[i]
        data_list = self.data_lists[i]
        decoded_list = [
            np.frombuffer(self.compressor.decode(data_list[idx]), dtype=dtype)
            for idx in h5_indices
        ]
        if len(decoded_list[0]) > self.chunk_size:
            decoded_list = [e.reshape(self.chunk_size, -1) for e in decoded_list]
        return decoded_list

    def _cov_norm(self, x):
        method = self.coverage_norm
        if method is None:
            return x
        elif method == "log":
            return np.log(x + 1.1)
        elif method == "linear":
            return x / 100
        else:
            raise ValueError(
                f"unrecognized method: {method} for normalizing coverage data"
            )

    def _update_sw_with_transition_weights(self):
        pass

    def _update_sw_with_coverage_weights(self):
        pass

    def _mk_timestep_pools(self, matrix):
        """reshape matrix to have multiple bp per timestep (in last dim)"""
        # assumes input shape
        # [0] = batch_size            --> don't touch
        # [1] = data's chunk_size     --> divide by pool size
        # [2:] = collapsable          --> -1, remaining, AKA np.prod(shape[2:]) * pool_size
        pool_size = self.model.pool_size
        if matrix is None:
            return None
        shape = list(matrix.shape)
        shape[1] = shape[1] // pool_size
        shape[-1] = -1
        matrix = matrix.reshape((shape))
        return matrix

    def _mk_timestep_pools_class_last(self, matrix):
        """reshape matrix to have multiple bp per timestep, w/ classes as last dim for softmax"""
        if matrix is None:
            return None
        pool_size = self.model.pool_size
        assert len(matrix.shape) == 3
        # assumes input shape
        # [0] = batch_size            --> don't touch
        # [1] = data's chunk_size     --> divide by pool size
        # [2] = labels                --> pooling inserted before, retained as last dimension
        matrix = matrix.reshape(
            (
                matrix.shape[0],
                matrix.shape[1] // pool_size,
                pool_size,  # make labels 2d so we can use the standard softmax / loss functions
                matrix.shape[-1],
            )
        )
        return matrix

    def _aggregate_timestep_pools(self, matrix, aggr_function=np.mean):
        pass

    def compress_tw(self, transitions):
        return self._squish_tw_to_sw(
            transitions, self.transition_weights, self.stretch_transition_weights
        )

    @staticmethod
    def _squish_tw_to_sw(transitions, transition_weights, stretch):
        # basically completes a max pool of transitions, from their pre-timestep pooled form
        max_pooled_trns = np.max(transitions, axis=2).astype(np.int8)

        # multiply by weights, and sum classes
        pooled_weighted_trns = np.multiply(max_pooled_trns, transition_weights)
        summed_trns = np.sum(pooled_weighted_trns, axis=2)

        # replace 0's (where weighting shouldn't change) with 1's as this will be multiplied later
        summed_trns[summed_trns == 0] = 1

        # maybe stretch transitions to upweight surrounding region somewhat (DEPRECATED bc performance was equivalent)
        if stretch != 0:
            summed_trns = HelixerSequence._apply_stretch(summed_trns, stretch)

        return summed_trns

    @staticmethod
    def _apply_stretch(reshaped_sw_t, stretch):
        """modifies sample weight shaped transitions so that they are a peak instead of a single point"""
        reshaped_sw_t = np.array(reshaped_sw_t)
        dilated_rf = np.ones(np.shape(reshaped_sw_t))

        where = np.where(reshaped_sw_t > 1)
        i = np.array(where[0])  # i unchanged
        j = np.array(where[1])  # j +/- step

        # find dividers depending on the size of the dilated rf
        dividers = []
        for distance in range(1, stretch + 1):
            dividers.append(2**distance)

        for z in range(stretch, 0, -1):
            dilated_rf[i, np.maximum(np.subtract(j, z), 0)] = np.maximum(
                reshaped_sw_t[i, j] / dividers[z - 1], 1
            )
            dilated_rf[i, np.minimum(np.add(j, z), len(dilated_rf[0]) - 1)] = (
                np.maximum(reshaped_sw_t[i, j] / dividers[z - 1], 1)
            )
        dilated_rf[i, j] = np.maximum(reshaped_sw_t[i, j], 1)
        return dilated_rf

    def __len__(self):
        """how many batches in epoch"""
        if self.debug:
            # if self.debug and self.mode == 'train':
            return 3
        elif self.overlap:
            return self.ol_helper.adjusted_epoch_length()
        else:
            return int(np.ceil(self.n_seqs / self.batch_size))

    @abstractmethod
    def __getitem__(self, idx):
        pass

    def _generic_get_item(self, idx):
        """covers the data preprocessing (reshape, trim, weighting, etc) common to all models"""
        X, y, sw, transitions, phases, _, coverage_scores = self._get_batch_data(idx)
        pool_size = self.model.pool_size

        if pool_size > 1:
            if X.shape[1] % pool_size != 0:
                # clip to maximum size possible with the pooling length
                overhang = X.shape[1] % pool_size
                X = X[:, :-overhang]
                if not self.only_predictions:
                    y = y[:, :-overhang]
                    sw = sw[:, :-overhang]
                    if self.predict_phase:
                        phases = phases[:, :-overhang]
                    if self.mode == "train" and self.transition_weights is not None:
                        transitions = transitions[:, :-overhang]

            if not self.only_predictions:
                y = self._mk_timestep_pools_class_last(y)
                sw = sw.reshape((sw.shape[0], -1, pool_size))
                sw = np.logical_not(np.any(sw == 0, axis=2)).astype(np.int8)

            if self.mode == "train":
                if self.class_weights is not None:
                    # class weights are additive for the individual timestep predictions
                    # giving even more weight to transition points
                    # class weights without pooling not supported yet
                    # cw = np.array([1.0, 1.2, 1.0, 0.8], dtype=np.float32)
                    cls_arrays = [
                        np.any((y[:, :, :, col] == 1), axis=2) for col in range(4)
                    ]
                    cls_arrays = np.stack(cls_arrays, axis=2).astype(np.int8)
                    # add class weights to applicable timesteps
                    cw_arrays = np.multiply(
                        cls_arrays, np.tile(self.class_weights, y.shape[:2] + (1,))
                    )
                    cw = np.sum(cw_arrays, axis=2)
                    sw = np.multiply(cw, sw)

                # todo, while now compressed, the following is still 1:1 with LSTM model... --> HelixerModel
                if self.transition_weights is not None:
                    transitions = self._mk_timestep_pools_class_last(transitions)
                    # more reshaping and summing  up transition weights for multiplying with sample weights
                    sw_t = self.compress_tw(transitions)
                    sw = np.multiply(sw_t, sw)

                if self.coverage_weights:
                    coverage_scores = coverage_scores.reshape(
                        (coverage_scores.shape[0], -1, pool_size)
                    )
                    # maybe offset coverage scores [0,1] by small number (bc RNAseq has issues too), default 0.0
                    if self.coverage_offset > 0.0:
                        coverage_scores = np.add(coverage_scores, self.coverage_offset)
                    coverage_scores = np.mean(coverage_scores, axis=2)
                    sw = np.multiply(coverage_scores, sw)

            if self.predict_phase and not self.only_predictions:
                y_phase = self._mk_timestep_pools_class_last(phases)
                y = [y, y_phase]

            return X, y, sw, transitions, phases, _, coverage_scores


class HelixerModel(ABC):
    def __init__(self, cli_args=None):
        self.cli_args = (
            cli_args  # if cli_args is None, the parameters from sys.argv will be used
        )

        self.parser = argparse.ArgumentParser()
        self.parser.add_argument("-d", "--data-dir", type=str, default=None)
        self.parser.add_argument(
            "-s", "--save-model-path", type=str, default="./best_model.h5"
        )
        self.parser.add_argument(
            "--large-eval-folder", type=str, default="", help=argparse.SUPPRESS
        )
        # training params
        self.parser.add_argument("-e", "--epochs", type=int, default=10000)
        self.parser.add_argument("-b", "--batch-size", type=int, default=8)
        self.parser.add_argument("--val-test-batch-size", type=int, default=32)
        self.parser.add_argument("--loss", type=str, default="")
        self.parser.add_argument("--patience", type=int, default=3)
        self.parser.add_argument("--check-every-nth-batch", type=int, default=1_000_000)
        self.parser.add_argument("--optimizer", type=str, default="adamw")
        self.parser.add_argument("--clip-norm", type=float, default=3.0)
        self.parser.add_argument("--learning-rate", type=float, default=3e-4)
        self.parser.add_argument("--weight-decay", type=float, default=3.5e-5)
        self.parser.add_argument("--class-weights", type=str, default="None")
        self.parser.add_argument("--transition-weights", type=str, default="None")
        self.parser.add_argument(
            "--stretch-transition-weights", type=int, default=0, help=argparse.SUPPRESS
        )  # bc no clear effect on result quality
        self.parser.add_argument(
            "--coverage-weights", action="store_true", help=argparse.SUPPRESS
        )
        self.parser.add_argument(
            "--coverage-offset", type=float, default=0.0, help=argparse.SUPPRESS
        )
        self.parser.add_argument(
            "--calculate-uncertainty", action="store_true", help=argparse.SUPPRESS
        )
        self.parser.add_argument(
            "--no-utrs", action="store_true", help=argparse.SUPPRESS
        )  # is this still needed?
        self.parser.add_argument("--predict-phase", action="store_true")
        self.parser.add_argument(
            "--load-predictions", action="store_true", help=argparse.SUPPRESS
        )  # bc no models that can use this are available
        self.parser.add_argument("--resume-training", action="store_true")
        # testing / predicting
        self.parser.add_argument("-l", "--load-model-path", type=str, default="")
        self.parser.add_argument("-t", "--test-data", type=str, default="")
        self.parser.add_argument(
            "-p", "--prediction-output-path", type=str, default="predictions.h5"
        )
        self.parser.add_argument(
            "--compression",
            default="gzip",
            help="compression used for datasets in predictions "
            'h5 file. One of "lzf" or "gzip" (default)',
        )
        self.parser.add_argument("--eval", action="store_true")
        self.parser.add_argument(
            "--overlap",
            action="store_true",
            help="will improve prediction quality at 'chunk' ends by creating and overlapping "
            "sliding-window predictions (with proportional increase in time usage)",
        )
        self.parser.add_argument(
            "--overlap-offset",
            type=int,
            default=None,
            help="distance to 'step' between predicting subsequences when overlapping "
            "(default: subsequence_length / 2)",
        )
        self.parser.add_argument(
            "--core-length",
            type=int,
            default=None,
            help="length of 'core' subsequence to retain before overlapping. The ends beyond this"
            "region will be cropped. (default: subsequence_length * 3 / 4)",
        )
        # resources
        self.parser.add_argument("--float-precision", type=str, default="float32")
        self.parser.add_argument("--cpus", type=int, default=1)
        self.parser.add_argument(
            "--gpu-id",
            type=int,
            default=-1,
            help="sets GPU index, use if you want to train on one GPU on a multi-GPU machine "
            "without a job scheduler system",
        )
        self.parser.add_argument(
            "--workers",
            type=int,
            default=1,
            help="umber of threads used to fetch input data; "
            "consider setting to match the number of GPUs",
        )
        # misc flags
        self.parser.add_argument("--save-every-check", action="store_true")
        self.parser.add_argument("--nni", action="store_true")
        self.parser.add_argument("-v", "--verbose", action="store_true")
        self.parser.add_argument("--debug", action="store_true")
        tuner = self.parser.add_argument_group(
            "fine tuning",
            "experimental parameters for training (a) final layer(s) "
            "of the model on target species (or other small dataset) "
            "with or without coverage, and "
            "with the rest of the model weights locked to reduce over fitting",
        )
        tuner.add_argument(
            "--fine-tune",
            action="store_true",
            help="use with --resume-training to replace and fine tune just the very last layer",
        )
        tuner.add_argument(
            "--pretrained-model-path",
            help="required when predicting with a model fine tuned with coverage; hopefully temporary",
        )
        tuner.add_argument(
            "--input-coverage",
            action="store_true",
            help='use "evaluation/rnaseq_(spliced_)coverage" from h5 as additional input '
            "for a late layer of the model",
        )
        tuner.add_argument(
            "--coverage-norm",
            default=None,
            help="None, linear or log (recommended); how coverage will be normalized before inputting",
        )
        tuner.add_argument(
            "--post-coverage-hidden-layer",
            action="store_true",
            help="adds extra dense layer between concatenating coverage and final output layer",
        )

        self.coverage_count = None

    def parse_args(self):
        """Parses the arguments either from the command line via argparse by using self.parser or
        takes a list of cli arguments from self.cli_args. This can be used to invoke a HelixerModel from
        another script."""
        args = vars(self.parser.parse_args(args=self.cli_args))
        self.__dict__.update(args)

        self.testing = bool(self.load_model_path and not self.resume_training)
        self.only_predictions = (
            self.testing and not self.eval
        )  # do only load X in this case

        if not self.testing:
            assert self.data_dir is not None, "--data-dir required for training"

        assert not (not self.testing and self.test_data)
        assert not (
            self.resume_training and (not self.load_model_path or not self.data_dir)
        )

        self.class_weights = eval(self.class_weights)
        if not isinstance(self.class_weights, (list, np.ndarray, type(None))):
            raise ValueError(
                f"--class-weights evaluated to {self.class_weights} of type {type(self.class_weights)}; "
                f"this commonly means you need to remove nested quotes if not starting with nni"
            )
        if type(self.class_weights) is list:
            self.class_weights = np.array(self.class_weights, dtype=np.float32)

        self.transition_weights = eval(self.transition_weights)
        if not isinstance(self.transition_weights, (list, np.ndarray, type(None))):
            raise ValueError(
                f"--transition-weights evaluated to {self.transition_weights} "
                f"of type {type(self.transition_weights)}; "
                f"this commonly means you need to remove nested quotes if not starting with nni"
            )
        if type(self.transition_weights) is list:
            self.transition_weights = np.array(
                self.transition_weights, dtype=np.float32
            )

        if self.verbose:
            print("HelixerModel config:")
            pprint(args)

    def generate_callbacks(self, train_generator):
        callbacks = [
            ConfusionMatrixTrain(
                self.save_model_path,
                train_generator,
                self.gen_validation_data(),
                self.large_eval_folder,
                self.patience,
                calc_H=self.calculate_uncertainty,
                report_to_nni=self.nni,
                check_every_nth_batch=self.check_every_nth_batch,
                save_every_check=self.save_every_check,
            ),
            PreshuffleCallback(train_generator),
        ]
        return callbacks

    def set_resources(self):
        gpu_devices = tf.config.experimental.list_physical_devices("GPU")
        for device in gpu_devices:
            tf.config.experimental.set_memory_growth(device, True)

        K.set_floatx(self.float_precision)
        if self.gpu_id > -1:
            tf.config.set_visible_devices([gpu_devices[self.gpu_id]], "GPU")

    def gen_training_data(self):
        SequenceCls = self.sequence_cls()
        return SequenceCls(
            model=self,
            h5_files=self.h5_trains,
            mode="train",
            batch_size=self.batch_size,
            shuffle=True,
        )

    def gen_validation_data(self):
        SequenceCls = self.sequence_cls()
        return SequenceCls(
            model=self,
            h5_files=self.h5_vals,
            mode="val",
            batch_size=self.val_test_batch_size,
            shuffle=False,
        )

    def gen_test_data(self):
        SequenceCls = self.sequence_cls()
        return SequenceCls(
            model=self,
            h5_files=self.h5_tests,
            mode="test",
            batch_size=self.val_test_batch_size,
            shuffle=False,
        )

    @staticmethod
    def run_metrics(generator, model, print_to_stdout=True, calc_H=False):
        start = time.time()
        metrics_calculator = Metrics(
            generator, print_to_stdout=print_to_stdout, skip_uncertainty=not calc_H
        )
        metrics = metrics_calculator.calculate_metrics(model)
        genic_metrics = metrics["genic_base_wise"]["genic"]
        if np.isnan(genic_metrics["f1"]):
            genic_metrics["f1"] = 0.0
        print(
            "\nmetrics calculation took: {:.2f} minutes\n".format(
                int(time.time() - start) / 60
            )
        )
        return genic_metrics["precision"], genic_metrics["recall"], genic_metrics["f1"]

    @staticmethod
    def run_large_eval(
        folder, model, generator, training_species, print_to_stdout=False, calc_H=False
    ):
        def print_table(results, table_name, training_species):
            table = [["Name", "Precision", "Recall", "F1-Score"]]
            for name, values in results:
                if name.lower() in training_species:
                    name += " (T)"
                table.append([name] + [f"{v:.4f}" for v in values])
            print(table)

        results = []
        training_species = [s.lower() for s in training_species]
        eval_file_names = glob.glob(f"{folder}/*.h5")
        for i, eval_file_name in enumerate(eval_file_names):
            h5_eval = h5py.File(eval_file_name, "r")
            species_name = os.path.basename(eval_file_name).split(".")[0]
            print(
                f"\nEvaluating with a sample of {species_name} ({i + 1}/{len(eval_file_names)})"
            )

            # possibly adjust batch size based on sample length, which could be flexible
            # assume the given batch size is for 20k length
            sample_len = h5_eval["data/X"].shape[1]
            adjusted_batch_size = int(generator.batch_size * (20000 / sample_len))
            print(f"adjusted batch size is {adjusted_batch_size}")

            # use exactly the data generator that is used during validation
            GenCls = generator.__class__
            gen = GenCls(
                model=generator.model,
                h5_file=h5_eval,
                mode="val",
                batch_size=adjusted_batch_size,
                shuffle=False,
            )
            perf_one_species = HelixerModel.run_metrics(
                gen, model, print_to_stdout=print_to_stdout, calc_H=calc_H
            )
            results.append([species_name, perf_one_species])
        # print results in tables sorted alphabetically and by f1
        results_by_name = sorted(results, key=lambda r: r[0])
        results_by_f1 = sorted(results, key=lambda r: r[1][2], reverse=True)
        print_table(
            results_by_name, "Generalization Validation by Name", training_species
        )
        print_table(
            results_by_f1, "Generalization Validation by Genic F1", training_species
        )

        # print one number summaries
        f1_scores = np.array([r[1][2] for r in results])
        in_train = np.array(
            [r[0].lower() in training_species for r in results], dtype=np.bool
        )
        table = [["Metric", "All", "Training", "Evaluation"]]
        for name, func in zip(
            ["Median F1", "Average F1", "Stddev F1"], [np.median, np.mean, np.std]
        ):
            table.append(
                [
                    name,
                    f"{func(f1_scores):.4f}",
                    f"{func(f1_scores[in_train]):.4f}",
                    f"{func(f1_scores[~in_train]):.4f}",
                ]
            )
        print(table)
        return np.median(f1_scores[~in_train])

    @abstractmethod
    def sequence_cls(self):
        pass

    @abstractmethod
    def model(self):
        pass

    @abstractmethod
    def model_hat(self, penultimate_layers):
        pass

    @abstractmethod
    def compile_model(self, model):
        pass

    @staticmethod
    def plot_model(model):
        from tensorflow.keras.utils import plot_model

        plot_model(model, to_file="model.png")
        print("Plotted to model.png")
        sys.exit()

    @staticmethod
    def sum_shapes(datasets):
        shapes = [ds.shape for ds in datasets]
        return [sum(x[0] for x in shapes)] + list(shapes[0][1:])

    def open_data_files(self):
        def get_n_correct_seqs(h5_files):
            sum_n_correct = 0
            for h5_file in h5_files:
                if "err_samples" in h5_file["/data"].keys():
                    err_samples = np.array(h5_file["/data/err_samples"])
                    n_correct = np.count_nonzero(err_samples is False)
                    if n_correct == 0:
                        print("WARNING: no fully correct sample found")
                else:
                    n_correct = 0
                sum_n_correct += n_correct
            return sum_n_correct

        def get_n_intergenic_seqs(h5_files):
            sum_n_fully_ig = 0
            for h5_file in h5_files:
                if "fully_intergenic_samples" in h5_file["/data"].keys():
                    ic_samples = np.array(h5_file["/data/fully_intergenic_samples"])
                    n_fully_ig = np.count_nonzero(ic_samples is True)
                    if n_fully_ig == 0:
                        print("WARNING: no fully intergenic samples found")
                else:
                    n_fully_ig = 0
                sum_n_fully_ig += n_fully_ig
            return sum_n_fully_ig

        if not self.testing:
            self.h5_trains = [
                h5py.File(f, "r")
                for f in glob.glob(os.path.join(self.data_dir, "training_data*h5"))
            ]
            self.h5_vals = [
                h5py.File(f, "r")
                for f in glob.glob(os.path.join(self.data_dir, "validation_data*h5"))
            ]
            try:
                self.shape_train = self.sum_shapes(
                    [h5["/data/X"] for h5 in self.h5_trains]
                )
            except IndexError as e:
                print(
                    "debugging info: self.h5_trains = {}, self.data_dir = {}".format(
                        self.h5_trains, self.data_dir
                    ),
                    file=sys.stderr,
                )
                raise e
            try:
                self.shape_val = self.sum_shapes([h5["/data/X"] for h5 in self.h5_vals])
            except IndexError as e:
                print(
                    "debugging info: self.h5_vals = {}, self.data_dir = {}".format(
                        self.h5_vals, self.data_dir
                    ),
                    file=sys.stderr,
                )
                raise e

            n_train_correct_seqs = get_n_correct_seqs(self.h5_trains)
            n_val_correct_seqs = get_n_correct_seqs(self.h5_vals)

            n_train_seqs = self.shape_train[0]
            n_val_seqs = self.shape_val[0]  # always validate on all

            n_intergenic_train_seqs = get_n_intergenic_seqs(self.h5_trains)
            n_intergenic_val_seqs = get_n_intergenic_seqs(self.h5_vals)
        else:
            self.h5_tests = [
                h5py.File(self.test_data, "r")
            ]  # list for consistency with train/val
            self.shape_test = self.h5_tests[0]["/data/X"].shape

            n_test_correct_seqs = get_n_correct_seqs(self.h5_tests)
            n_test_seqs_with_intergenic = self.shape_test[0]
            n_intergenic_test_seqs = get_n_intergenic_seqs(self.h5_tests)

        if self.verbose:
            print("\nData config: ")
            if not self.testing:
                print([dict(x.attrs) for x in self.h5_trains])
                print("\nTraining data/X shape: {}".format(self.shape_train[:2]))
                print("Validation data/X shape: {}".format(self.shape_val[:2]))
                print("\nTotal est. training sequences: {}".format(n_train_seqs))
                print("Total est. val sequences: {}".format(n_val_seqs))
                print(
                    "\nEst. intergenic train/val seqs: {:.2f}% / {:.2f}%".format(
                        n_intergenic_train_seqs / n_train_seqs * 100,
                        n_intergenic_val_seqs / n_val_seqs * 100,
                    )
                )
                print(
                    "Fully correct train/val seqs: {:.2f}% / {:.2f}%\n".format(
                        n_train_correct_seqs / self.shape_train[0] * 100,
                        n_val_correct_seqs / self.shape_val[0] * 100,
                    )
                )
            else:
                print([dict(x.attrs) for x in self.h5_tests])
                print("\nTest data shape: {}".format(self.shape_test[:2]))
                print(
                    "\nIntergenic test seqs: {:.2f}%".format(
                        n_intergenic_test_seqs / n_test_seqs_with_intergenic * 100
                    )
                )
                print(
                    "Fully correct test seqs: {:.2f}%\n".format(
                        n_test_correct_seqs / self.shape_test[0] * 100
                    )
                )

    def _make_predictions(self, model):
        # loop through batches and continuously expand output dataset as everything might
        # not fit in memory
        pred_out = h5py.File(self.prediction_output_path, "w")
        test_sequence = self.gen_test_data()

        for batch_index in range(len(test_sequence)):
            if self.verbose:
                print(batch_index, "/", len(test_sequence), end="\r")
            if not self.only_predictions:
                input_data = test_sequence[batch_index][0]
            else:
                input_data = test_sequence[batch_index]
            try:
                predictions = model.predict_on_batch(input_data)
            except Exception as e:
                print(
                    "Errors at prediction often result from exhausting the GPU RAM."
                    "Your RAM requirement depends on subsequence_length x (val_test_)batch_size."
                    "That and the network size (can be changed during training but not inference)."
                )
                raise e
            if isinstance(predictions, list):
                # when we have two outputs, one is for phase
                output_names = ["predictions", "predictions_phase"]
            else:
                # if we just had one output
                predictions = (predictions,)
                output_names = ["predictions"]

            for dset_name, pred_dset in zip(output_names, predictions):
                # join last two dims when predicting one hot labels
                pred_dset = pred_dset.reshape(pred_dset.shape[:2] + (-1,))
                # reshape when predicting more than one point at a time
                label_dim = 4
                if pred_dset.shape[2] != label_dim:
                    n_points = pred_dset.shape[2] // label_dim
                    pred_dset = pred_dset.reshape(
                        pred_dset.shape[0],
                        pred_dset.shape[1] * n_points,
                        label_dim,
                    )
                    # add 0-padding if needed
                    n_removed = self.shape_test[1] - pred_dset.shape[1]
                    if n_removed > 0:
                        zero_padding = np.zeros(
                            (pred_dset.shape[0], n_removed, pred_dset.shape[2]),
                            dtype=pred_dset.dtype,
                        )
                        pred_dset = np.concatenate((pred_dset, zero_padding), axis=1)
                else:
                    n_removed = 0  # just to avoid crashing with Unbound Local Error setting attrs for dCNN

                if self.overlap:
                    pred_dset = test_sequence.ol_helper.overlap_predictions(
                        batch_index, pred_dset
                    )

                # prepare h5 dataset and save the predictions to disk
                pred_dset = pred_dset.astype(np.float16)
                if batch_index == 0:
                    old_len = 0
                    pred_out.create_dataset(
                        dset_name,
                        data=pred_dset,
                        maxshape=(None,) + pred_dset.shape[1:],
                        chunks=(1,) + pred_dset.shape[1:],
                        dtype="float16",
                        compression=self.compression,
                        shuffle=True,
                    )
                else:
                    old_len = pred_out[dset_name].shape[0]
                    pred_out[dset_name].resize(old_len + pred_dset.shape[0], axis=0)
                pred_out[dset_name][old_len:] = pred_dset

        # add model config and other attributes to predictions
        h5_model = h5py.File(self.load_model_path, "r")
        pred_out.attrs["model_config"] = h5_model.attrs["model_config"]
        pred_out.attrs["n_bases_removed"] = n_removed
        pred_out.attrs["test_data_path"] = self.test_data
        pred_out.attrs["model_path"] = self.load_model_path
        pred_out.attrs["timestamp"] = str(datetime.datetime.now())
        if hasattr(self, "loaded_model_hash"):
            pred_out.attrs["model_md5sum"] = self.loaded_model_hash
        pred_out.close()
        h5_model.close()

    def _print_model_info(self, model):
        pwd = os.getcwd()
        os.chdir(os.path.dirname(__file__))
        try:
            if os.path.isfile(self.load_model_path):
                cmd = ["md5sum", self.load_model_path]
                self.loaded_model_hash = subprocess.check_output(cmd).strip().decode()
        except subprocess.CalledProcessError:
            print(
                "An error occurred while running a subprocess, unable to record loaded_model_hash"
            )
            self.loaded_model_hash = "error"

        if self.verbose:
            print(model.summary())
        os.chdir(pwd)  # return to previous directory

    def run(self):
        def load_model_strategy():
            if not self.input_coverage:
                model = load_model(self.load_model_path)
            else:
                # for whatever reason, the fine tuning method is not saving the full model
                # in an entirely valid h5 file (depending on if you ask h5py or h5ls). puh.
                # thus loading the original model is both the easiest way to get architecture
                # setup and seems safer to make sure _all_ and not just _new_ weights are there
                oldmodel = load_model(self.pretrained_model_path)
                # repeat everything done setting up training to get exact architecture
                # freeze weights and replace everything from the dense layer
                dense_at = [l.name for l in oldmodel.layers].index("dense")
                for layer in oldmodel.layers:
                    layer.trainable = False

                model = self.insert_coverage_before_hat(oldmodel, dense_at)
                model.load_weights(self.load_model_path)
            return model

        self.set_resources()
        self.open_data_files()
        if self.input_coverage:
            # preview first h5 file to find number of bam files and set 'coverage_count'
            # which is used to calculate model input size +
            try:
                h5_files = self.h5_tests
            except AttributeError:
                h5_files = self.h5_trains
            self.coverage_count = h5_files[0]["evaluation/rnaseq_coverage"].shape[2]

        # we're training, not eval nor predict
        if not self.testing:
            if self.resume_training:
                if not self.fine_tune:
                    model = load_model(self.load_model_path)
                else:
                    oldmodel = load_model(self.load_model_path)
                    assert oldmodel.input.shape[2] == 4, (
                        f"input shape of trained model != 4 ({oldmodel.input.shape[2]}); "
                        f"fine tuning is only supported on models trained without coverage"
                    )
                    # freeze weights and replace everything from the dense layer
                    dense_at = [l.name for l in oldmodel.layers].index("dense")
                    for layer in oldmodel.layers:
                        layer.trainable = False
                    # the following assumes the base-model is trained without coverage
                    if not self.input_coverage:
                        inp = oldmodel.input
                        output = self.model_hat(
                            (oldmodel.layers[dense_at - 1].output, None)
                        )
                        model = Model(inp, output)
                    else:
                        model = self.insert_coverage_before_hat(oldmodel, dense_at)

            else:
                model = self.model()
            self._print_model_info(model)

            if self.optimizer.lower() == "adam":
                self.optimizer = optimizers.Adam(
                    learning_rate=self.learning_rate, clipnorm=self.clip_norm
                )
            elif self.optimizer.lower() == "adamw":
                self.optimizer = AdamW(
                    learning_rate=self.learning_rate,
                    clipnorm=self.clip_norm,
                    weight_decay=self.weight_decay,
                )

            self.compile_model(model)

            train_generator = self.gen_training_data()
            model.fit(
                train_generator,
                epochs=self.epochs,
                workers=self.workers,
                callbacks=self.generate_callbacks(train_generator),
                verbose=True,
            )
        else:
            assert self.test_data.endswith(
                ".h5"
            ), "Need a h5 test data file when loading a model"
            assert self.load_model_path.endswith(".h5"), "Need a h5 model file"

            strategy = tf.distribute.MirroredStrategy()
            if strategy.num_replicas_in_sync > 1:
                with strategy.scope():
                    model = load_model_strategy()
            else:
                # use no strategy if there is no replication to avoid warnings from the dataset not being tf.data
                model = load_model_strategy()

            self._print_model_info(model)

            if self.eval:
                test_generator = self.gen_test_data()
                _, _, _ = HelixerModel.run_metrics(
                    test_generator, model, calc_H=self.calculate_uncertainty
                )
                if self.large_eval_folder:
                    assert (
                        self.data_dir != ""
                    ), "need training data of the model for training genome names"
                    training_species = h5py.File(
                        os.path.join(self.data_dir, "training_data.h5"), "r"
                    ).attrs["genomes"]
                    _ = HelixerModel.run_large_eval(
                        self.large_eval_folder,
                        model,
                        test_generator,
                        training_species,
                        print_to_stdout=True,
                        calc_H=self.calculate_uncertainty,
                    )
            else:
                if os.path.isfile(self.prediction_output_path):
                    print(
                        f"WARNING: {self.prediction_output_path} already exists and will be overwritten."
                    )
                self._make_predictions(model)
            for h5_test in self.h5_tests:
                h5_test.close()

    def insert_coverage_before_hat(self, oldmodel, dense_at):
        """splits input in half, feeds CATG to the main model, and coverage in before tuning layers"""
        # hacking RNAseq coverage in.
        values_per_bp = 4 + self.coverage_count * 2

        raw_input = Input(
            shape=(None, values_per_bp), dtype=self.float_precision, name="main_input"
        )
        # make a callable model that gives the intermediate output, with first 4 of input (CATG)
        excerpt_model = Model(oldmodel.input, oldmodel.layers[dense_at - 2].output)
        x = excerpt_model(raw_input[:, :, :4])
        # add hat, including coverage back on
        output = self.model_hat((x, raw_input[:, :, 4:]))

        model = Model(raw_input, output)
        return model



import numpy as np


# some helpers for handling / sorting / or checking sort of our h5 files
def mk_seqonly_keys(h5):
    return [a + b for a, b in zip(h5["data/species"], h5["data/seqids"])]


def mk_keys(h5, flip=False):
    first_idx = 0
    second_idx = 1
    if flip:
        first_idx, second_idx = second_idx, first_idx
    return zip(
        h5["data/species"],
        h5["data/seqids"],
        h5["data/start_ends"][:, first_idx],
        h5["data/start_ends"][:, second_idx],
    )


def get_sp_seq_ranges(h5):
    # dict with {sp: {"start: N,
    #                 "end": N,
    #                 "seqids": {seqid: [start, end], seqid2: [start2, end2], ...}},
    #            sp2: {...},
    #            ...}
    out = {}

    gen = zip(range(h5["data/y"].shape[0]), h5["data/species"], h5["data/seqids"])

    i, prev_sp, prev_seqid = next(gen)
    out[prev_sp] = {
        "start": i,  # 0
        "seqids": {prev_seqid: [i]},
    }
    for i, sp, seqid in gen:
        if sp != prev_sp:
            # end previous sp and seqid
            out[prev_sp]["end"] = i
            out[prev_sp]["seqids"][prev_seqid].append(i)
            # open new
            out[sp] = {"start": i, "seqids": {seqid: [i]}}
        elif seqid != prev_seqid:
            # end previous seqid
            out[sp]["seqids"][prev_seqid].append(i)
            # open new
            out[sp]["seqids"][seqid] = [i]
        prev_sp, prev_seqid = sp, seqid
    # end final sp/seqid
    out[prev_sp]["end"] = i + 1
    out[prev_sp]["seqids"][prev_seqid].append(i + 1)
    return out


# additional helping functions for predictions to hints, here so they can be tested
# also probably some redundancy with above to clean up (-_-)
def get_contiguous_ranges(h5):
    """gets h5 coordinates for same species, sequence and strand AKA end to end across a chromosome/scaffold"""
    start_ends = h5["data/start_ends"][:]
    marks_unique = np.stack(
        (h5["data/seqids"], h5["data/species"], start_ends[:, 1] > start_ends[:, 0])
    )
    items, indexes, lengths = np.unique(
        marks_unique, axis=1, return_index=True, return_counts=True
    )
    reindex = np.argsort(indexes)
    for i in reindex:
        seqid, species, is_plus_strand = items[:, i]
        if is_plus_strand == b"True":
            bool_strand = True
        elif is_plus_strand == b"False":
            bool_strand = False
        else:
            raise ValueError(
                "is_plus_strand somehow not True/False, instead it is {}".format(
                    is_plus_strand
                )
            )
        yield {
            "species": species,
            "seqid": seqid,
            "is_plus_strand": bool_strand,
            "start_i": indexes[i],
            "end_i": indexes[i] + lengths[i],
        }


def read_in_chunks(preds, data, start_i, end_i, step=100):
    for i in range(start_i, end_i, step):
        ei = min(i + step, end_i)
        pred_chunk = preds["predictions"][i:ei]
        pred_chunk = pred_chunk.reshape((-1, 4))
        start = data["data/start_ends"][i, 0]
        end = data["data/start_ends"][ei - 1, 1]
        # in case of padding
        if abs(end - start) % data["data/X"].shape[1]:
            # padded areas have no sequence, so will use that to create a mask for simplicity
            mask = data["data/X"][i:ei].reshape((-1, 4))
            mask = np.any(mask, axis=1)
            pred_chunk = pred_chunk[mask]
            assert pred_chunk.shape[0] == abs(start - end)
        yield pred_chunk, start, end


def find_confident_single_class_regions(pred_chunk, pad=5):
    """find start, end (relative to pred_chunk) bits where in-between is confidently one class"""
    # in the event of a seamless class swap [1,0,0,0] -> [0,1,0,0]
    # and of a thin-seemed swap [1,0,0,0] -> [0.5,0.5,0,0] -> [0,1,0,0]
    # (and any easier swap, as well)
    # the 2-bp smoothing below will allow us to detect a class swap by drop in confidence (max)
    shift_n_averaged = (pred_chunk[:-1] + pred_chunk[1:]) / 2
    # find anywhere network was not confident / or class switch
    lowconf_idx = np.where(np.max(shift_n_averaged, axis=1) < 0.75001)[0]
    if lowconf_idx.shape[0] == 0:
        yield 0, pred_chunk.shape[0]
        return
    # handle (& yield) start of chunk edge case if pred_chunk starts with a confident region
    if lowconf_idx[0] > pad * 2:
        yield 0, lowconf_idx[0]
    # invert above to find longer single-class confident regions (double our padding)
    dist2next_lowconf = lowconf_idx[1:] - lowconf_idx[:-1]
    conf_sub_idx = np.where(dist2next_lowconf > pad * 2)[0]

    # yield confident regions
    for sub_idx in conf_sub_idx:
        start = lowconf_idx[sub_idx] + 1  # confidence starts _after_ lowconf ends
        end = (
            start + dist2next_lowconf[sub_idx] - 1
        )  # (exclusive) confidence ends where lowconf starts
        yield start, end

    # handle (& yield) end of chunk edge case if pred_chunk ends with a confident region
    if pred_chunk.shape[0] - lowconf_idx[-1] > pad * 2:
        yield lowconf_idx[-1] + 1, pred_chunk.shape[0]


def divvy_by_confidence(one_class_chunk, step_key, pad=5, stability_threshold=0.1):
    """breaks down contiguous 1-class region to pre-hints with semi-consistent confidence"""
    main_class = np.argmax(one_class_chunk[0])
    min_step, max_size = step_key[main_class]
    diffs = np.abs(one_class_chunk[:-1, main_class] - one_class_chunk[1:, main_class])
    cumulative_diffs = np.cumsum(diffs)
    # we start after padding (and end before), not necessarily at 0 nor sequence end
    cdiff_at_last_yield = cumulative_diffs[pad]
    end_of_last_yield = pad
    padded_seq_end = one_class_chunk.shape[0] - pad
    for i in range(pad, padded_seq_end, min_step):
        end = min(i + min_step, padded_seq_end)
        # if the total observed diffs since yield have passed the threshold, yield again
        # thus where confidence is volatile one gets small chunks, and for continuous predictions
        # one gets large (probably max_size) chunks (saves gff size & there's little gain in breaking down)
        if (
            cumulative_diffs[end] - cdiff_at_last_yield > stability_threshold
            or end - end_of_last_yield > max_size
            or end == padded_seq_end
        ):
            start = end_of_last_yield
            yield {
                "category": main_class,
                "start": start,
                "end": end,
                "confidence": np.mean(one_class_chunk[start:end, main_class]),
            }
            # reset trackers
            end_of_last_yield = end
            cdiff_at_last_yield = cumulative_diffs[end]


def file_stem(path):
    """Returns the file name without extension"""
    import os

    return os.path.basename(path).split(".")[0]



import time
import h5py
import pyfastx
import datetime
from .numerify import CoordNumerifier


class HelixerExportControllerBase(object):
    def __init__(self, input_path, output_path, match_existing=False):
        self.input_path = input_path
        self.output_path = output_path
        self.match_existing = match_existing

    @staticmethod
    def calc_n_chunks(coord_len, chunk_size):
        """calculates the number of chunks resulting from a coord len and chunk size"""
        n_chunks = coord_len // chunk_size
        if coord_len % chunk_size:
            n_chunks += 1  # bc pad to size
        n_chunks *= 2  # for + & - strand
        return n_chunks

    @staticmethod
    def _create_dataset(
        h5_file, key, matrix, dtype, compression="gzip", create_empty=True
    ):
        shape = list(matrix.shape)
        shuffle = len(shape) > 1
        if create_empty:
            shape[0] = 0  # create w/o size
        h5_file.create_dataset(
            key,
            shape=shape,
            maxshape=tuple([None] + shape[1:]),
            chunks=tuple([1] + shape[1:]),
            dtype=dtype,
            compression=compression,
            shuffle=shuffle,
        )  # only for the compression

    def _create_or_expand_datasets(
        self, h5_group, flat_data, n_chunks, compression="gzip"
    ):
        if h5_group not in self.h5 or len(self.h5[h5_group].keys()) == 0:
            for mat_info in flat_data:
                self._create_dataset(
                    self.h5,
                    h5_group + mat_info.key,
                    mat_info.matrix,
                    mat_info.dtype,
                    compression,
                )

        old_len = self.h5[h5_group + flat_data[0].key].shape[0]
        self.h5_coord_offset = old_len
        for mat_info in flat_data:
            self.h5[h5_group + mat_info.key].resize(old_len + n_chunks, axis=0)

    def _save_data(
        self,
        flat_data,
        h5_coords,
        n_chunks,
        first_round_for_coordinate,
        compression="gzip",
        h5_group="/data/",
    ):
        assert (
            len(set(mat_info.matrix.shape[0] for mat_info in flat_data)) == 1
        ), "unequal data lengths"

        if first_round_for_coordinate:
            self._create_or_expand_datasets(h5_group, flat_data, n_chunks, compression)

        # h5_coords are relative for the coordinate/chromosome, so offset by previous length
        old_len = self.h5_coord_offset
        start = old_len + h5_coords[0]
        end = old_len + h5_coords[1]

        # writing to the h5 file
        for mat_info in flat_data:
            self.h5[h5_group + mat_info.key][start:end] = mat_info.matrix
        self.h5.flush()

    def _add_data_attrs(self):
        attrs = {
            "timestamp": str(datetime.datetime.now()),
            "input_path": self.input_path,
        }
        # insert attrs into .h5 file
        for key, value in attrs.items():
            self.h5.attrs[key] = value


class HelixerFastaToH5Controller(HelixerExportControllerBase):
    class CoordinateSurrogate(object):
        """Mimics some functionality of the Coordinate orm class, so we can go directly from FASTA to H5"""

        def __init__(self, seqid, seq):
            self.seqid = seqid
            self.sequence = seq
            self.length = len(seq)

        def __repr__(self):
            return f"Fasta only Coordinate (seqid: {self.seqid}, len: {self.length})"

    def export_fasta_to_h5(
        self, chunk_size, compression, multiprocess, species, write_by
    ):
        assert write_by >= chunk_size, (
            "when specifying '--write-by' it needs to be larger than "
            "or equal to '--subsequence-length'"
        )
        self.h5 = h5py.File(self.output_path, "w")

        for i, (seqid, seq) in enumerate(
            pyfastx.Fasta(self.input_path, build_index=False)
        ):
            coord = HelixerFastaToH5Controller.CoordinateSurrogate(seqid, seq)
            n_chunks = HelixerExportControllerBase.calc_n_chunks(
                coord.length, chunk_size
            )
            data_gen = CoordNumerifier.numerify_only_fasta(
                coord,
                chunk_size,
                species,
                use_multiprocess=multiprocess,
                write_by=write_by,
            )
            for j, strand_res in enumerate(data_gen):
                data, h5_coords = strand_res
                self._save_data(
                    data,
                    h5_coords=h5_coords,
                    n_chunks=n_chunks,
                    first_round_for_coordinate=(j == 0),
                    compression=compression,
                )
        self._add_data_attrs()
        self.h5.close()



#!/usr/bin/env python3

import pytest
import os
import tempfile
import h5py
import numpy as np


@pytest.fixture
def temp_fasta_file():
    """Create a temporary FASTA file for testing."""
    with tempfile.NamedTemporaryFile(suffix='.fasta', mode='w+') as fasta_file:
        fasta_file.write(">seq1\nACGT\n>seq2\nTGCA\n")
        fasta_file.flush()
        yield fasta_file.name


@pytest.fixture
def temp_h5_file():
    """Create a temporary HDF5 file for testing."""
    with tempfile.NamedTemporaryFile(suffix='.h5') as h5_file:
        yield h5_file.name


@pytest.fixture
def temp_gff_file():
    """Create a temporary GFF3 file for testing."""
    with tempfile.NamedTemporaryFile(suffix='.gff3') as gff_file:
        yield gff_file.name


@pytest.fixture
def mock_h5_data():
    """Create mock HDF5 data for testing."""
    with tempfile.NamedTemporaryFile(suffix='.h5') as h5_file:
        with h5py.File(h5_file.name, 'w') as f:
            # Create datasets
            f.create_dataset('/data/X', data=np.zeros((10, 4)))
            f.create_dataset('/data/species', data=np.array(['test_species'] * 10, dtype='S20'))
            f.create_dataset('/data/seqids', data=np.array(['seq1'] * 10, dtype='S20'))
            f.create_dataset('/data/start_ends', data=np.zeros((10, 2), dtype=np.int32))
            
            # Set attributes
            f.attrs['timestamp'] = 'test_timestamp'
            f.attrs['input_path'] = 'test_input_path'
        
        yield h5_file.name



#! /usr/bin/env python3

"""handler code for sliding window predictions and re-overlapping things back to original sequences"""

import numpy as np
import math
import sys


def _n_ori_chunks_from_batch_chunks(max_batch_size, overlap_offset, chunk_size):
    """check max number of original (non overlapped) chunks that fit in overlapped batch_size (or remaining)"""
    # this is going to be empirical for A simplicity and B maintainability
    end = 0
    empirical_bs = 0
    while empirical_bs <= max_batch_size:
        end += 1
        sb = SubBatch(
            tuple(range(end)),
            edge_handle_start=False,
            edge_handle_end=False,
            is_plus_strand=True,
            overlap_offset=overlap_offset,
            chunk_size=chunk_size,
        )
        empirical_bs = sb.sub_batch_size

    return end - 1  # the while loop only breaks _after_ overshooting max


class SubBatch:
    def __init__(
        self,
        h5_indices,
        edge_handle_start,
        edge_handle_end,
        is_plus_strand,
        overlap_offset,
        chunk_size,
        keep_start=None,
        keep_end=None,
    ):
        self.h5_indices = h5_indices
        # the following parameters are primarily for debugging
        self.keep_start = keep_start
        self.keep_end = keep_end
        # if not on a sequence edge start and end bits will be cropped
        # also, start on negative strand would ideally have special handling for weird padding
        self.edge_handle_start = edge_handle_start
        self.edge_handle_end = edge_handle_end
        self.is_plus_strand = is_plus_strand
        self.overlap_depth = math.ceil(chunk_size / overlap_offset)
        self.overlap_offset = overlap_offset
        self.chunk_size = chunk_size
        self.sliding_coordinates = self._mk_sliding_coordinates()
        self.sub_batch_size = len(self.sliding_coordinates)

    def __repr__(self):
        sstr, estr = "(", ")"

        if self.edge_handle_end:
            estr = "]"
        if self.edge_handle_start:
            sstr = "["

        return "SubBatch, {}edges{}, h5 indices: {}".format(sstr, estr, self.h5_indices)

    @property
    def seq_length(self):
        return self.chunk_size * len(self.h5_indices)

    def _mk_sliding_coordinates(self):
        out = []
        for i in range(0, self.seq_length - self.chunk_size + 1, self.overlap_offset):
            out.append((i, i + self.chunk_size))
        # add one final one for special case with uneven fit
        # [chunkCHUNK] with offset 3 characters
        # [00000     ]
        #     11111  ]
        #        2222]2   <--- not this X
        #       22222]    <--- but this :-)
        if out[-1][1] < self.seq_length:
            out.append((self.seq_length - self.chunk_size, self.seq_length))
        return tuple(out)

    def mk_sliding_overlaps_for_data_sub_batch(self, data_sub_batch):
        """makes sliding window of input data (x, or coverage data)"""
        # combine first 2 dimensions (i.e. merge chunks)
        dat = data_sub_batch.reshape(
            [np.prod(data_sub_batch.shape[:2])] + list(data_sub_batch.shape[2:])
        )
        sliding_dat = [dat[start:end] for start, end in self.sliding_coordinates]
        return sliding_dat

    def _overlap_preds(self, preds, core_length):
        """take sliding-window predictions, and overlap (w/end clipping) to generate original coordinate predictions"""
        trim_by = (self.chunk_size - core_length) // 2
        ydim = preds[0].shape[-1]
        if ydim == self.chunk_size:
            ydim = 1
        preds_out = np.zeros(shape=(self.seq_length, ydim))
        counts = np.zeros(shape=(self.seq_length, 1))

        len_preds = len(preds)
        for i, chunk, start_end in zip(
            range(len_preds), preds, self.sliding_coordinates
        ):
            start, end = start_end
            # cut to core, (but not sequence ends)
            if trim_by > 0:
                if i > 0:  # all except first seq
                    start += trim_by
                    chunk = chunk[trim_by:]
                if i < len_preds - 1:  # all except last seq
                    end -= trim_by
                    chunk = chunk[:-trim_by]
            elif trim_by < 0:  # sanity check only
                raise ValueError(
                    "invalid trim value: {}. Maybe core_length {} > chunk_size {}?".format(
                        trim_by, core_length, chunk.shape[0]
                    )
                )
            sub_counts = counts[start:end]
            # average weighted by number of predictions counted at position so far
            preds_out[start:end] = (preds_out[start:end] * sub_counts + chunk) / (
                sub_counts + 1
            )
            # increment counted so far
            counts[start:end] += 1
        preds_out = preds_out.reshape((len(self.h5_indices), self.chunk_size, ydim))
        return preds_out

    def overlap_and_edge_handle_preds(self, preds, core_length):
        """overlaps sliding predictions, then crops as necessary on edges"""
        # the final sequences for what is cropped will come from previous/next batch instead
        # i.e. this should produce identical output regardless of batch size
        # as avoidable batch/sub-batch edge effects will be complete cropped here.
        clean_preds = self._overlap_preds(preds, core_length)
        clean_preds = self.edge_handle(clean_preds)
        return clean_preds

    def edge_handle(self, dat):
        """crops first and second array from output unless at sequence edge"""
        if not self.edge_handle_start:
            dat = dat[1:]
        if not self.edge_handle_end:
            dat = dat[:-1]
        return dat


# places where overlap will affect core functionality of HelixerSequence
class OverlapSeqHelper(object):
    """handles overlap-ready batching, as well as overlap-prep and overlapping there-of"""

    def __init__(
        self, contiguous_ranges, chunk_size, max_batch_size, overlap_offset, core_length
    ):
        # check validity of settings
        self.max_batch_size = max_batch_size
        self.core_length = core_length
        if chunk_size % overlap_offset:
            print(
                "chunk size is not divisible by overlap_offset, so coverage depth"
                "during overlapping will vary by 1",
                file=sys.stderr,
            )

        assert (
            overlap_offset <= core_length
        ), "change settings to over-, not under-lap (offset < core length)"
        assert core_length > 0
        assert overlap_offset > 0

        # contiguous ranges should be created by .helpers.get_contiguous_ranges
        self.sliding_batches = self._mk_sliding_batches(
            contiguous_ranges=contiguous_ranges,
            chunk_size=chunk_size,
            overlap_offset=overlap_offset,
        )

    def _mk_sliding_batches(self, contiguous_ranges, chunk_size, overlap_offset):
        # max_n_chunks is the number of chunks that will go into overlapping
        # before any sliding window, before any dropping or clipping
        max_n_chunks = _n_ori_chunks_from_batch_chunks(
            self.max_batch_size, overlap_offset, chunk_size
        )
        # make sure we can drop first and last (to produce same results, regardless of batch size)
        # and still have one chunk
        assert max_n_chunks >= 3, (
            "batch_size is set too small to functionally overlap, "
            "either a) set higher (to ~ 2 * chunk_size / overlap_offset + 1), "
            "b) increase overlap_offset, or c) don't overlap"
        )
        step = max_n_chunks - 2  # -2 bc ends will be cropped
        # most of these will effectively be final batches, but short seqs/ends may be grouped together (for efficiency)
        sub_batches = []

        for crange in contiguous_ranges:
            # step through sequence so that non-edges can have 1-chunk cropped off start/end
            # and regenerate original sequence with a simple concatenation there after
            for i in range(crange["start_i"], crange["end_i"], step):
                sub_batch_start = max(
                    i - 1, crange["start_i"]
                )  # pad 1 left (except seq edge)
                keep_start = max(i, crange["start_i"])
                keep_end = min(i + step, crange["end_i"])
                sub_batch_end = min(
                    i + step + 1, crange["end_i"]
                )  # pad 1 right (except seq edge)
                h5_indices = tuple(range(sub_batch_start, sub_batch_end))
                sub_batches.append(
                    SubBatch(
                        h5_indices,
                        is_plus_strand=crange["is_plus_strand"],
                        edge_handle_start=i == crange["start_i"],
                        edge_handle_end=i + step + 1 > crange["end_i"],
                        keep_start=keep_start,
                        keep_end=keep_end,
                        overlap_offset=overlap_offset,
                        chunk_size=chunk_size,
                    )
                )

        # group into final batches, so as to keep total size <= max_batch_size
        # i.e. achieve consistent (& user adjustable) memory usage on graphics card
        sliding_batches = []
        batch = []
        batch_total_size = 0
        for sb in sub_batches:
            if batch_total_size + sb.sub_batch_size <= self.max_batch_size:
                batch.append(sb)
                batch_total_size += sb.sub_batch_size
            else:
                sliding_batches.append(batch)
                batch = [sb]
                batch_total_size = sb.sub_batch_size
        sliding_batches.append(batch)
        return sliding_batches

    def adjusted_epoch_length(self):
        """number of batches per epoch (given that we're overlapping)"""
        return len(self.sliding_batches)

    def h5_indices_of_batch(self, batch_idx):
        """concatenate indices from sub batches to give all indices for the batch at {batch_idx}"""
        sub_batches = self.sliding_batches[batch_idx]
        h5_indices = []
        for sb in sub_batches:
            h5_indices += sb.h5_indices
        return np.array(h5_indices)

    def make_input(self, batch_idx, data_batch):
        """make sliding input for prediction and overlapping (i.e. for X, maybe also for coverage)"""
        sub_batches = self.sliding_batches[batch_idx]
        sb_input_lengths = [len(sb.h5_indices) for sb in sub_batches]
        sb_input_starts = np.cumsum(sb_input_lengths) - sb_input_lengths
        x_as_list = []
        for start, length, sb in zip(sb_input_starts, sb_input_lengths, sub_batches):
            x_as_list.append(
                sb.mk_sliding_overlaps_for_data_sub_batch(
                    data_batch[start : (start + length)]
                )
            )
        sliding_input = np.concatenate(x_as_list)
        return sliding_input

    def overlap_predictions(self, batch_idx, predictions):
        """overlapping of sliding predictions to regenerate original dimensions"""
        sub_batches = self.sliding_batches[batch_idx]
        sub_batch_lengths = [sb.sub_batch_size for sb in sub_batches]
        sub_batch_starts = np.cumsum(sub_batch_lengths) - sub_batch_lengths
        out = []
        for start, length, sb in zip(sub_batch_starts, sub_batch_lengths, sub_batches):
            preds = predictions[start : (start + length)]
            out.append(sb.overlap_and_edge_handle_preds(preds, self.core_length))
        out = np.concatenate(out)
        n_expect = np.sum([sb.keep_end - sb.keep_start for sb in sub_batches])
        assert out.shape[0] == n_expect, (
            f"overlapping trouble, maybe try a higher batch-size? "
            f"(debug info: got {out.shape[0]}, expected {n_expect},"
            f"sub batches: {sub_batches})"
        )
        return out

    def subset_input(self, batch_idx, y_true_or_sw):
        """generate subset from data corresponding to _final_ predictions, i.e. to run y_true through during eval"""
        sub_batches = self.sliding_batches[batch_idx]
        sb_input_lengths = [len(sb.h5_indices) for sb in sub_batches]
        sb_input_starts = np.cumsum(sb_input_lengths) - sb_input_lengths
        dat_as_list = []
        for start, length, sb in zip(sb_input_starts, sb_input_lengths, sub_batches):
            dat_as_list.append(sb.edge_handle(y_true_or_sw[start : (start + length)]))
        return np.concatenate(dat_as_list)



#!/usr/bin/env python3

import pytest
from unittest.mock import patch, MagicMock

from helixerlite.hybrid_model import HybridModel


class TestHybridModel:
    """Tests for the HybridModel class."""

    @pytest.mark.skip(reason="Requires complex mocking of argparse")
    @patch("helixerlite.hybrid_model.HelixerModel")
    def test_init(self, mock_helixer_model):
        """Test the initialization of HybridModel."""
        # Create a mock instance with a parser attribute
        mock_instance = MagicMock()
        mock_helixer_model.return_value = mock_instance
        mock_instance.parser = MagicMock()

        # Call the constructor with test arguments
        args = ["--test-arg", "test_value"]
        HybridModel(args)  # No need to store the model instance

        # Assertions
        mock_helixer_model.assert_called_once_with(cli_args=args)
        # Check that add_argument was called for the expected arguments
        assert mock_instance.parser.add_argument.call_count >= 6

    @pytest.mark.skip(reason="Requires complex mocking of argparse")
    @patch("helixerlite.hybrid_model.HelixerModel")
    def test_run(self, mock_helixer_model):
        """Test the run method of HybridModel."""
        # Create a mock instance with necessary attributes
        mock_instance = MagicMock()
        mock_helixer_model.return_value = mock_instance
        mock_instance.parser = MagicMock()
        mock_instance.predict_and_eval = MagicMock()

        # Set up the mock instance to have the necessary attributes to pass validation
        mock_instance.testing = False
        mock_instance.data_dir = "/path/to/data"

        # Call the constructor and run method
        hybrid_model = HybridModel([])
        hybrid_model.run()

        # Assertions
        mock_instance.predict_and_eval.assert_called_once()



#!/usr/bin/env python3

import pytest
import os
import tempfile
import h5py
import numpy as np
from unittest.mock import patch, MagicMock

from helixerlite.hdf5 import HelixerFastaToH5Controller, HelixerExportControllerBase


class TestHelixerExportControllerBase:
    """Tests for the HelixerExportControllerBase class."""

    def test_calc_n_chunks(self):
        """Test the calc_n_chunks static method."""
        # Test with exact division
        assert HelixerExportControllerBase.calc_n_chunks(100, 10) == 20  # (100 // 10) * 2
        
        # Test with remainder
        assert HelixerExportControllerBase.calc_n_chunks(105, 10) == 22  # ((105 // 10) + 1) * 2
        
        # Test with chunk size larger than coord length
        assert HelixerExportControllerBase.calc_n_chunks(5, 10) == 2  # (5 // 10 + 1) * 2

    @pytest.fixture
    def controller(self):
        """Create a controller instance for testing."""
        with tempfile.NamedTemporaryFile(suffix='.fasta') as input_file, \
             tempfile.NamedTemporaryFile(suffix='.h5') as output_file:
            controller = HelixerExportControllerBase(input_file.name, output_file.name)
            yield controller

    def test_init(self, controller):
        """Test the initialization of HelixerExportControllerBase."""
        assert hasattr(controller, 'input_path')
        assert hasattr(controller, 'output_path')
        assert hasattr(controller, 'match_existing')
        assert controller.match_existing is False


class TestHelixerFastaToH5Controller:
    """Tests for the HelixerFastaToH5Controller class."""

    @pytest.fixture
    def mock_fasta_file(self):
        """Create a temporary FASTA file for testing."""
        with tempfile.NamedTemporaryFile(suffix='.fasta', mode='w+') as fasta_file:
            fasta_file.write(">seq1\\nACGT\\n>seq2\\nTGCA\\n")
            fasta_file.flush()
            yield fasta_file.name

    @pytest.fixture
    def mock_h5_file(self):
        """Create a temporary HDF5 file for testing."""
        with tempfile.NamedTemporaryFile(suffix='.h5') as h5_file:
            yield h5_file.name

    def test_coordinate_surrogate(self):
        """Test the CoordinateSurrogate inner class."""
        surrogate = HelixerFastaToH5Controller.CoordinateSurrogate("seq1", "ACGT")
        assert surrogate.seqid == "seq1"
        assert surrogate.sequence == "ACGT"
        assert surrogate.length == 4
        assert "seq1" in str(surrogate)
        assert "len: 4" in str(surrogate)

    @patch('helixerlite.hdf5.pyfastx.Fasta')
    @patch('helixerlite.hdf5.h5py.File')
    @patch('helixerlite.hdf5.CoordNumerifier.numerify_only_fasta')
    def test_export_fasta_to_h5(self, mock_numerify, mock_h5py_file, mock_pyfastx, 
                               mock_fasta_file, mock_h5_file):
        """Test the export_fasta_to_h5 method."""
        # Setup mocks
        mock_h5_file_instance = MagicMock()
        mock_h5py_file.return_value = mock_h5_file_instance
        
        mock_pyfastx.return_value = [
            ("seq1", "ACGT"),
            ("seq2", "TGCA")
        ]
        
        # Mock the numerify_only_fasta method to return test data
        mock_data = MagicMock()
        mock_data.key = "test_key"
        mock_data.matrix = np.zeros((10, 4))
        mock_data.dtype = np.float32
        
        mock_numerify.return_value = [
            ([mock_data], (0, 10)),  # First strand
            ([mock_data], (10, 20))  # Second strand
        ]
        
        # Create controller and call the method
        controller = HelixerFastaToH5Controller(mock_fasta_file, mock_h5_file)
        controller.h5 = mock_h5_file_instance  # Set the h5 attribute directly
        
        # Add a mock for _save_data
        controller._save_data = MagicMock()
        
        # Call the method
        controller.export_fasta_to_h5(
            chunk_size=10,
            compression="gzip",
            multiprocess=True,
            species="test_species",
            write_by=100
        )
        
        # Assertions
        mock_pyfastx.assert_called_once_with(mock_fasta_file, build_index=False)
        mock_h5py_file.assert_called_once_with(mock_h5_file, "w")
        
        # Check that _save_data was called for each sequence and strand
        assert controller._save_data.call_count == 4  # 2 sequences * 2 strands
        
        # Check that the file was closed
        mock_h5_file_instance.close.assert_called_once()

    @pytest.mark.skipif(not os.path.exists('tests/genome.fasta'),
                       reason="Test FASTA file not available")
    def test_integration_with_real_file(self, mock_h5_file):
        """Integration test with a real FASTA file."""
        try:
            # Create controller
            controller = HelixerFastaToH5Controller('tests/genome.fasta', mock_h5_file)
            
            # Call the method with small values to make the test faster
            controller.export_fasta_to_h5(
                chunk_size=10,
                compression="gzip",
                multiprocess=False,  # Use False for testing
                species="test_species",
                write_by=100
            )
            
            # Check that the output file exists and is not empty
            assert os.path.exists(mock_h5_file)
            assert os.path.getsize(mock_h5_file) > 0
            
            # Open the file and check its structure
            with h5py.File(mock_h5_file, 'r') as h5:
                # Check that the expected datasets exist
                assert '/data/X' in h5
                assert '/data/species' in h5
                assert '/data/seqids' in h5
                
                # Check that the attributes were set
                assert 'timestamp' in h5.attrs
                assert 'input_path' in h5.attrs
                assert h5.attrs['input_path'] == 'tests/genome.fasta'
                
        except Exception as e:
            pytest.skip(f"Error in integration test: {str(e)}")



#!/usr/bin/env python3

# No need for pytest or os imports
import tempfile
from unittest.mock import patch, MagicMock

from helixerlite.log import startLogging, system_info, finishLogging


class TestLog:
    """Tests for the log module."""

    def test_start_logging(self):
        """Test the startLogging function."""
        # Mock tracemalloc to avoid starting it
        with (
            patch("helixerlite.log.tracemalloc") as mock_tracemalloc,
            patch("helixerlite.log.logging") as mock_logging,
            patch("helixerlite.log.os.path.isfile") as mock_isfile,
            patch("helixerlite.log.os.remove") as mock_remove,
        ):
            # Setup mocks
            mock_logger = MagicMock()
            mock_logging.getLogger.return_value = mock_logger
            mock_formatter = MagicMock()
            mock_logging.Formatter.return_value = mock_formatter
            mock_handler = MagicMock()
            mock_logging.FileHandler.return_value = mock_handler
            mock_console_handler = MagicMock()
            mock_logging.StreamHandler.return_value = mock_console_handler
            mock_isfile.return_value = True  # Simulate file exists

            # Call the function
            with tempfile.NamedTemporaryFile(suffix=".log") as temp_log:
                startLogging(logfile=temp_log.name)  # No need to store the logger

                # Assertions
                mock_tracemalloc.start.assert_called_once()
                mock_logging.getLogger.assert_called_once()
                mock_isfile.assert_called_once_with(temp_log.name)
                mock_remove.assert_called_once_with(temp_log.name)
                mock_logging.FileHandler.assert_called_once_with(temp_log.name)
                assert (
                    mock_logger.addHandler.call_count >= 2
                )  # At least 2 handlers added

    def test_system_info(self):
        """Test the system_info function."""
        # Mock all the imported modules
        with (
            patch("platform.python_version") as mock_python_version,
            patch("helixerlite.log.__version__", "0.1.0"),
            patch("helixerlite.log.tf.__version__", "2.10.0"),
            patch("helixerlite.log.h5py.__version__", "3.7.0"),
            patch("helixerlite.log.pyfastx.__version__", "0.8.4"),
            patch("helixerlite.log.gfftk.__version__", "0.2.0"),
            patch("helixerlite.log.numpy.__version__", "1.23.5"),
        ):
            # Setup mocks
            mock_python_version.return_value = "3.10.0"

            # Create a mock logger function
            mock_logger_func = MagicMock()

            # Call the function
            system_info(mock_logger_func)

            # Assertions
            mock_logger_func.assert_called_once()
            mock_python_version.assert_called_once()

            # Check that the log message contains the version information
            log_message = mock_logger_func.call_args[0][0]
            assert "Python v3.10.0" in log_message
            assert "helixerlite v0.1.0" in log_message
            assert "tensorflow v2.10.0" in log_message
            assert "h5py v3.7.0" in log_message
            assert "pyfastx v0.8.4" in log_message
            assert "gfftk v0.2.0" in log_message
            assert "numpy v1.23.5" in log_message

    def test_finish_logging(self):
        """Test the finishLogging function."""
        # Mock tracemalloc
        with (
            patch("helixerlite.log.tracemalloc") as mock_tracemalloc,
            patch("helixerlite.log.human_readable_size") as mock_human_readable_size,
        ):
            # Setup mocks
            mock_tracemalloc.get_traced_memory.return_value = (
                1000,
                5000,
            )  # current, peak
            mock_human_readable_size.return_value = "5.0 KB"

            # Create a mock logger function
            mock_logger_func = MagicMock()

            # Call the function
            finishLogging(mock_logger_func, "test_module")

            # Assertions
            mock_tracemalloc.get_traced_memory.assert_called_once()
            mock_tracemalloc.stop.assert_called_once()
            mock_human_readable_size.assert_called_once_with(5000)
            mock_logger_func.assert_called_once()

            # Check the log message
            log_message = mock_logger_func.call_args[0][0]
            assert "test_module" in log_message
            assert "finished" in log_message
            assert "peak memory usage=5.0 KB" in log_message



import subprocess
import sys
import os
import errno
import socket
from urllib.request import urlopen
import helixerpost
from .__init__ import __version__


def human_readable_size(size, decimal_places=2):
    """
    Convert a size in bytes to a human-readable string representation.

    This function takes a size in bytes and converts it to a more human-readable format,
    using units such as B, KiB, MiB, GiB, TiB, and PiB. The size is formatted to the specified
    number of decimal places.

    Parameters:
    - size (float): The size in bytes to convert.
    - decimal_places (int, optional): The number of decimal places to include in the output (default is 2).

    Returns:
    - str: A string representing the size in a human-readable format with the appropriate unit.
    """
    for unit in ["B", "KiB", "MiB", "GiB", "TiB", "PiB"]:
        if size < 1024.0 or unit == "PiB":
            break
        size /= 1024.0
    return f"{size:.{decimal_places}f} {unit}"


def download(url, name, wget=False):
    """
    Download a file from a given URL.

    This function downloads a file from the specified URL and saves it with the given name.
    It can use either the `wget` command-line tool or Python's `urllib` for downloading,
    depending on the `wget` flag.

    Parameters:
    - url (str): The URL of the file to download.
    - name (str): The name to save the downloaded file as.
    - wget (bool, optional): Flag to use `wget` for downloading (default is False).

    Returns:
    None
    """
    if wget:
        # download with wget
        cmd = ["wget", "-O", name, "--no-check-certificate", "-t", "2", "-c", url]
        subprocess.call(cmd)
    else:
        file_name = name
        try:
            u = urlopen(url)
            f = open(file_name, "wb")
            block_sz = 8192
            while True:
                buffer = u.read(block_sz)
                if not buffer:
                    break
                f.write(buffer)
            f.close()
        except socket.error as e:
            if e.errno != errno.ECONNRESET:
                raise
            pass


def prediction2gff3(
    hd_genome,
    hd_prediction,
    gffout,
    window_size=100,
    edge_threshold=0.1,
    peak_threshold=0.8,
    min_coding_length=60,
):
    # helixer_post_bin <genome.h5> <predictions.h5> <window_size> <edge_threshold> <peak_threshold> <min_coding_length> <output.gff3>
    cmd = [
        "helixer_post_bin",
        hd_genome,
        hd_prediction,
        str(window_size),
        str(edge_threshold),
        str(peak_threshold),
        str(min_coding_length),
        gffout,
    ]
    runprocess(cmd)


def preds2gff3(
    hd_genome,
    hd_prediction,
    gffout,
    window_size=100,
    edge_threshold=0.1,
    peak_threshold=0.8,
    min_coding_length=60,
):
    # helixer_post_bin <genome.h5> <predictions.h5> <window_size> <edge_threshold> <peak_threshold> <min_coding_length> <output.gff3>
    helixerpost.run_helixer_post(
        hd_genome,
        hd_prediction,
        window_size,
        edge_threshold,
        peak_threshold,
        min_coding_length,
        gffout,
    )


def runprocess(cmd, stdout=False, stderr=False, cwd=".", debug=False):
    if not stdout and not stderr:
        proc = subprocess.Popen(
            cmd, cwd=cwd, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
    elif stdout and not stderr:
        with open(stdout, "w") as outfile:
            proc = subprocess.Popen(
                cmd, cwd=cwd, stdout=outfile, stderr=subprocess.PIPE
            )
    elif not stdout and stderr:
        with open(stderr, "w") as outfile:
            proc = subprocess.Popen(
                cmd, cwd=cwd, stdout=subprocess.PIPE, stderr=outfile
            )
    elif stdout and stderr:
        if stdout == stderr:
            with open(stdout, "w") as outfile:
                proc = subprocess.Popen(cmd, cwd=cwd, stdout=outfile, stderr=outfile)
        else:
            with open(stdout, "w") as outfile1:
                with open(stderr, "w") as outfile2:
                    proc = subprocess.Popen(
                        cmd, cwd=cwd, stdout=outfile1, stderr=outfile2
                    )
    stdout, stderr = proc.communicate()
    if proc.returncode != 0:
        sys.stderr.write("CMD ERROR: {}".format(" ".join(cmd)))
        if stdout:
            sys.stderr.write(stdout.decode("utf-8"))
        if stderr:
            sys.stderr.write(stderr.decode("utf-8"))
        sys.exit(1)
    if debug:
        if stdout:
            sys.stderr.write(stdout.decode("utf-8"))
        if stderr:
            sys.stderr.write(stderr.decode("utf-8"))


def execute(cmd):
    DEVNULL = open(os.devnull, "w")
    popen = subprocess.Popen(
        cmd, stdout=subprocess.PIPE, universal_newlines=True, stderr=DEVNULL
    )
    for stdout_line in iter(popen.stdout.readline, ""):
        yield stdout_line
    popen.stdout.close()
    return_code = popen.wait()
    if return_code:
        raise subprocess.CalledProcessError(return_code, cmd)


def execute_timeout(cmd, timeout=120):
    # stream execute but add a timeout
    DEVNULL = open(os.devnull, "w")
    try:
        p = subprocess.Popen(
            cmd, stdout=subprocess.PIPE, universal_newlines=True, stderr=DEVNULL
        )
        return_code = p.wait(timeout=timeout)
        if return_code:
            raise subprocess.CalledProcessError(return_code, cmd)
        else:
            for stdout_line in iter(p.stdout.readline, ""):
                yield stdout_line
            p.stdout.close()
    except subprocess.TimeoutExpired:
        # print(f"Timeout for {cmd} ({timeout}s) expired", file=sys.stderr)
        p.terminate()
        return ""


def check_inputs(inputs):
    for filename in inputs:
        if not is_file(filename):
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), filename)


def is_file(f):
    if os.path.isfile(f):
        return True
    else:
        return False


def which2(program):
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file
    return None


def open_pipe(command, mode="r", buff=1024 * 1024):
    import subprocess
    import signal

    if "r" in mode:
        return subprocess.Popen(
            command,
            shell=True,
            bufsize=buff,
            stdout=subprocess.PIPE,
            universal_newlines=True,
            preexec_fn=lambda: signal.signal(signal.SIGPIPE, signal.SIG_DFL),
        ).stdout
    elif "w" in mode:
        return subprocess.Popen(
            command,
            shell=True,
            bufsize=buff,
            universal_newlines=True,
            stdin=subprocess.PIPE,
        ).stdin
    return None


NORMAL = 0
PROCESS = 1
PARALLEL = 2

WHICH_BZIP2 = which2("bzip2")
WHICH_PBZIP2 = which2("pbzip2")


def open_bz2(filename, mode="r", buff=1024 * 1024, external=PARALLEL):
    if external is None or external == NORMAL:
        import bz2

        return bz2.BZ2File(filename, mode, buff)
    elif external == PROCESS:
        if not WHICH_BZIP2:
            return open_bz2(filename, mode, buff, NORMAL)
        if "r" in mode:
            return open_pipe("bzip2 -dc " + filename, mode, buff)
        elif "w" in mode:
            return open_pipe("bzip2 >" + filename, mode, buff)
    elif external == PARALLEL:
        if not WHICH_PBZIP2:
            return open_bz2(filename, mode, buff, PROCESS)
        if "r" in mode:
            return open_pipe("pbzip2 -dc " + filename, mode, buff)
        elif "w" in mode:
            return open_pipe("pbzip2 >" + filename, mode, buff)
    return None


WHICH_GZIP = which2("gzip")
WHICH_PIGZ = which2("pigz")


def open_gz(filename, mode="r", buff=1024 * 1024, external=PARALLEL):
    if external is None or external == NORMAL:
        import gzip

        return gzip.GzipFile(filename, mode, buff)
    elif external == PROCESS:
        if not WHICH_GZIP:
            return open_gz(filename, mode, buff, NORMAL)
        if "r" in mode:
            return open_pipe("gzip -dc " + filename, mode, buff)
        elif "w" in mode:
            return open_pipe("gzip >" + filename, mode, buff)
    elif external == PARALLEL:
        if not WHICH_PIGZ:
            return open_gz(filename, mode, buff, PROCESS)
        if "r" in mode:
            return open_pipe("pigz -dc " + filename, mode, buff)
        elif "w" in mode:
            return open_pipe("pigz >" + filename, mode, buff)
    return None


WHICH_XZ = which2("xz")


def open_xz(filename, mode="r", buff=1024 * 1024, external=PARALLEL):
    if WHICH_XZ:
        if "r" in mode:
            return open_pipe("xz -dc " + filename, mode, buff)
        elif "w" in mode:
            return open_pipe("xz >" + filename, mode, buff)
    return None


def zopen(filename, mode="r", buff=1024 * 1024, external=PARALLEL):
    """
    Open pipe, zipped, or unzipped file automagically

    # external == 0: normal zip libraries
    # external == 1: (zcat, gzip) or (bzcat, bzip2)
    # external == 2: (pigz -dc, pigz) or (pbzip2 -dc, pbzip2)
    """
    if "r" in mode and "w" in mode:
        return None
    if filename.startswith("!"):
        return open_pipe(filename[1:], mode, buff)
    elif filename.endswith(".bz2"):
        return open_bz2(filename, mode, buff, external)
    elif filename.endswith(".gz"):
        return open_gz(filename, mode, buff, external)
    elif filename.endswith(".xz"):
        return open_xz(filename, mode, buff, external)
    else:
        return open(filename, mode, buff)
    return None



#!/usr/bin/env python3

from helixerlite.__main__ import fasta2hdf5

import pytest


class TestFasta2hdf5:
    """Tests for the fasta2hdf5 function in __main__.py."""

    def test_convert_valid_fasta_to_hdf5(self, mocker):
        """Test that fasta2hdf5 calls HelixerFastaToH5Controller with the correct parameters."""
        # Mock the HelixerFastaToH5Controller to avoid actual file operations
        mock_controller = mocker.patch(
            "helixerlite.__main__.HelixerFastaToH5Controller", autospec=True
        )
        mock_instance = mock_controller.return_value

        # Call the function
        fasta = "tests/genome.fasta"
        hdout = "tests/output.h5"
        fasta2hdf5(fasta, hdout)

        # Assert that the controller was created with the correct parameters
        mock_controller.assert_called_once_with(fasta, hdout)

        # Assert that the export_fasta_to_h5 method was called with the correct parameters
        mock_instance.export_fasta_to_h5.assert_called_once_with(
            chunk_size=21384,
            compression="gzip",
            multiprocess=True,
            species=mocker.ANY,
            write_by=20_000_000,
        )

    def test_convert_with_custom_parameters(self, mocker):
        """Test fasta2hdf5 with custom parameters."""
        # Mock the HelixerFastaToH5Controller
        mock_controller = mocker.patch(
            "helixerlite.__main__.HelixerFastaToH5Controller", autospec=True
        )
        mock_instance = mock_controller.return_value

        # Call the function with custom parameters
        fasta = "tests/genome.fasta"
        hdout = "tests/output.h5"
        subseqlen = 1000
        species = "test_species"
        fasta2hdf5(fasta, hdout, subseqlen=subseqlen, species=species)

        # Assert that the controller was created with the correct parameters
        mock_controller.assert_called_once_with(fasta, hdout)

        # Assert that the export_fasta_to_h5 method was called with the correct parameters
        mock_instance.export_fasta_to_h5.assert_called_once_with(
            chunk_size=subseqlen,
            compression="gzip",
            multiprocess=True,
            species=species,
            write_by=20_000_000,
        )



#!/usr/bin/env python3

import pytest
import os
import tempfile
import h5py
import numpy as np
from unittest.mock import patch, MagicMock

# Import the module to test
import helixerpost


class TestHelixerpost:
    """Tests for the helixerpost Rust module."""

    def test_import(self):
        """Test that the helixerpost module can be imported."""
        assert hasattr(helixerpost, "hello_world")
        assert hasattr(helixerpost, "run_helixer_post")

    def test_hello_world(self):
        """Test the hello_world function."""
        result = helixerpost.hello_world()
        assert result == "Hello, world!"
        assert isinstance(result, str)

    def test_run_helixer_post_signature(self):
        """Test that run_helixer_post has the expected signature."""
        import inspect

        sig = inspect.signature(helixerpost.run_helixer_post)
        params = list(sig.parameters.keys())

        # Check that the function has the expected parameters
        expected_params = [
            "genome_path",
            "predictions_path",
            "window_size",
            "edge_threshold",
            "peak_threshold",
            "min_coding_length",
            "gff_filename",
        ]
        assert params == expected_params

    def test_run_helixer_post_with_mock_data(self):
        """Test run_helixer_post with mock data files."""
        # Skip this test if the test data files don't exist
        if not os.path.exists("tests/data/genome_data.h5") or not os.path.exists(
            "tests/data/predictions.h5"
        ):
            pytest.skip("Test data files not available")

        with tempfile.NamedTemporaryFile(suffix=".gff3") as temp_gff:
            # Mock the run_helixer_post function
            original_run_helixer_post = helixerpost.run_helixer_post
            try:
                helixerpost.run_helixer_post = MagicMock(return_value=None)

                # Call the function with test data
                helixerpost.run_helixer_post(
                    "tests/data/genome_data.h5",
                    "tests/data/predictions.h5",
                    100,  # window_size
                    0.1,  # edge_threshold
                    0.8,  # peak_threshold
                    60,  # min_coding_length
                    temp_gff.name,
                )

                # Check that the mock was called with the correct arguments
                helixerpost.run_helixer_post.assert_called_once_with(
                    "tests/data/genome_data.h5",
                    "tests/data/predictions.h5",
                    100,
                    0.1,
                    0.8,
                    60,
                    temp_gff.name,
                )
            finally:
                # Restore the original function
                helixerpost.run_helixer_post = original_run_helixer_post

    @patch("helixerpost.run_helixer_post")
    def test_run_helixer_post_called_with_correct_args(self, mock_run_helixer_post):
        """Test that run_helixer_post is called with the correct arguments."""
        # Setup mock
        mock_run_helixer_post.return_value = None

        # Call the function
        helixerpost.run_helixer_post(
            "genome.h5", "predictions.h5", 100, 0.1, 0.8, 60, "output.gff3"
        )

        # Check that the function was called with the correct arguments
        mock_run_helixer_post.assert_called_once_with(
            "genome.h5", "predictions.h5", 100, 0.1, 0.8, 60, "output.gff3"
        )



#!/usr/bin/env python3

import pytest
import os
import tempfile
import subprocess
import h5py
import numpy as np
from unittest import mock
from unittest.mock import patch

from helixerlite.__main__ import fasta2hdf5
from helixerlite.utilities import preds2gff3


class TestIntegration:
    """Integration tests for the helixerlite package."""

    @pytest.mark.skipif(
        not os.path.exists("tests/genome.fasta"), reason="Test FASTA file not available"
    )
    def test_fasta2hdf5_integration(self):
        """Test the fasta2hdf5 function with a real FASTA file."""
        with tempfile.NamedTemporaryFile(suffix=".h5") as temp_h5:
            try:
                # Convert FASTA to HDF5
                fasta2hdf5("tests/genome.fasta", temp_h5.name, subseqlen=10)

                # Check that the output file exists and is not empty
                assert os.path.exists(temp_h5.name)
                assert os.path.getsize(temp_h5.name) > 0

                # Open the file and check its structure
                with h5py.File(temp_h5.name, "r") as h5:
                    # Check that the expected datasets exist
                    assert "/data/X" in h5
                    assert "/data/species" in h5
                    assert "/data/seqids" in h5

                    # Check that the attributes were set
                    assert "timestamp" in h5.attrs
                    assert "input_path" in h5.attrs
                    assert h5.attrs["input_path"] == "tests/genome.fasta"

            except Exception as e:
                pytest.skip(f"Error in fasta2hdf5 integration test: {str(e)}")

    @patch("helixerlite.utilities.helixerpost.run_helixer_post")
    def test_preds2gff3_integration(self, mock_run_helixer_post):
        """Test the preds2gff3 function with mock HDF5 files."""
        # Setup mock
        mock_run_helixer_post.return_value = None

        with tempfile.NamedTemporaryFile(suffix=".gff3") as temp_gff:
            # Convert predictions to GFF3
            preds2gff3(
                "tests/data/genome_data.h5",
                "tests/data/predictions.h5",
                temp_gff.name,
                window_size=100,
                edge_threshold=0.1,
                peak_threshold=0.8,
                min_coding_length=60,
            )

            # Check that run_helixer_post was called with the correct arguments
            mock_run_helixer_post.assert_called_once_with(
                "tests/data/genome_data.h5",
                "tests/data/predictions.h5",
                100,
                0.1,
                0.8,
                60,
                temp_gff.name,
            )

    @patch("helixerlite.hybrid_model.HybridModel")
    @patch("helixerlite.utilities.helixerpost.run_helixer_post")
    def test_full_pipeline_integration(self, mock_run_helixer_post, mock_hybrid_model):
        """Test the full pipeline from FASTA to GFF3."""
        # This test mocks the HybridModel part since it's complex and requires TensorFlow

        # Setup mocks
        mock_hybrid_model.return_value.run.return_value = None
        mock_run_helixer_post.return_value = None

        with tempfile.TemporaryDirectory() as temp_dir:
            # Setup file paths
            genome_h5 = os.path.join(temp_dir, "genome.h5")
            predictions_h5 = os.path.join(temp_dir, "predictions.h5")
            gff_output = os.path.join(temp_dir, "output.gff3")

            # Step 1: Convert FASTA to HDF5
            # Mock the HelixerFastaToH5Controller
            with patch(
                "helixerlite.__main__.HelixerFastaToH5Controller"
            ) as mock_controller:
                mock_instance = mock_controller.return_value

                fasta2hdf5("tests/data/genome.fasta", genome_h5, subseqlen=10)

                # Check that the controller was created with the correct parameters
                mock_controller.assert_called_once_with(
                    "tests/data/genome.fasta", genome_h5
                )

                # Check that export_fasta_to_h5 was called with the correct parameters
                mock_instance.export_fasta_to_h5.assert_called_once_with(
                    chunk_size=10,
                    compression="gzip",
                    multiprocess=True,
                    species=mock.ANY,
                    write_by=20_000_000,
                )

            # Step 2: Run the HybridModel
            # We don't need to do anything here since we've already mocked the HybridModel

            # Step 3: Convert predictions to GFF3
            preds2gff3(genome_h5, predictions_h5, gff_output)

            # Check that run_helixer_post was called with the correct arguments
            mock_run_helixer_post.assert_called_once_with(
                genome_h5,
                predictions_h5,
                100,  # default window_size
                0.1,  # default edge_threshold
                0.8,  # default peak_threshold
                60,  # default min_coding_length
                gff_output,
            )

    @pytest.mark.skipif(True, reason="This test requires a real model and is slow")
    def test_command_line_interface(self):
        """Test the command-line interface."""
        # This test is marked as skip by default because it's slow and requires a real model

        with tempfile.TemporaryDirectory() as temp_dir:
            output_gff = os.path.join(temp_dir, "output.gff3")

            # Run the command
            cmd = [
                "helixerlite",
                "--fasta",
                "tests/genome.fasta",
                "--out",
                output_gff,
                "--lineage",
                "fungi",
                "--cpus",
                "1",
            ]

            try:
                result = subprocess.run(cmd, capture_output=True, text=True)

                # Check that the command succeeded
                assert result.returncode == 0

                # Check that the output file exists
                assert os.path.exists(output_gff)
                assert os.path.getsize(output_gff) > 0

            except Exception as e:
                pytest.skip(f"Error in CLI test: {str(e)}")



import importlib.metadata
import sys

# Make helixerpost available at the package level
try:
    import helixerpost

    sys.modules["helixerlite.helixerpost"] = helixerpost
except ImportError as e:
    print(f"Error importing helixerpost module: {e}")

__version__ = importlib.metadata.version("helixerlite")



#!/usr/bin/env python3
import os
import sys
import zipfile
import tempfile
import subprocess
import glob

def main():
    # Find wheel files
    wheel_files = glob.glob("dist/*.whl")
    if not wheel_files:
        print("No wheel files found in dist/")
        return 1
    
    print(f"Found wheel files: {wheel_files}")
    
    # Create a temporary directory
    with tempfile.TemporaryDirectory() as temp_dir:
        for wheel_file in wheel_files:
            print(f"\nExamining wheel: {wheel_file}")
            
            # Extract the wheel
            with zipfile.ZipFile(wheel_file, 'r') as zip_ref:
                zip_ref.extractall(temp_dir)
                print(f"Extracted to: {temp_dir}")
                
                # List all files in the wheel
                print("\nFiles in the wheel:")
                for root, dirs, files in os.walk(temp_dir):
                    for file in files:
                        file_path = os.path.join(root, file)
                        rel_path = os.path.relpath(file_path, temp_dir)
                        print(f"  {rel_path}")
            
            # Install the wheel
            print("\nInstalling wheel...")
            result = subprocess.run(
                [sys.executable, "-m", "pip", "install", "--force-reinstall", wheel_file],
                capture_output=True,
                text=True
            )
            print(f"Installation stdout: {result.stdout}")
            print(f"Installation stderr: {result.stderr}")
            
            # Try to import the module
            print("\nTrying to import helixerlite...")
            try:
                import helixerlite
                print(f"Successfully imported helixerlite version {helixerlite.__version__}")
                print(f"helixerlite.__file__: {helixerlite.__file__}")
                print(f"helixerlite module contents: {dir(helixerlite)}")
                
                try:
                    import helixerpost
                    print(f"Successfully imported helixerpost")
                    print(f"helixerpost.__file__: {helixerpost.__file__}")
                    print(f"helixerpost module contents: {dir(helixerpost)}")
                except ImportError as e:
                    print(f"Error importing helixerpost: {e}")
            except ImportError as e:
                print(f"Error importing helixerlite: {e}")
    
    return 0

if __name__ == "__main__":
    sys.exit(main())



import os
import platform
import logging
import tracemalloc
from .__init__ import __version__
from .utilities import human_readable_size
import gfftk
import tensorflow as tf
import pyfastx
import h5py
import numpy


green = "\033[92m"
END = "\033[0m"


class CustomFormatter(logging.Formatter):
    green = "\033[92m"
    grey = "\x1b[38;20m"
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    end = "\033[0m"
    date_format = green + "%(asctime)s " + end
    msg_format = "%(message)s"
    file_format = "%(asctime)s %(message)s"

    FORMATS = {
        logging.DEBUG: date_format + grey + msg_format + reset,
        logging.INFO: date_format + grey + msg_format + reset,
        logging.WARNING: date_format + yellow + msg_format + reset,
        logging.ERROR: date_format + red + msg_format + reset,
        logging.CRITICAL: date_format + bold_red + msg_format + reset,
    }
    FILE_FORMATS = {
        logging.DEBUG: file_format,
        logging.INFO: file_format,
        logging.WARNING: file_format,
        logging.ERROR: file_format,
        logging.CRITICAL: file_format,
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt, datefmt="[%b %d %I:%M %p]")
        return formatter.format(record)


class FileFormatter(logging.Formatter):
    file_format = "%(asctime)s %(message)s"

    FILE_FORMATS = {
        logging.DEBUG: file_format,
        logging.INFO: file_format,
        logging.WARNING: file_format,
        logging.ERROR: file_format,
        logging.CRITICAL: file_format,
    }

    def format(self, record):
        log_fmt = self.FILE_FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt, datefmt="[%b %d %I:%M %p]")
        return formatter.format(record)


def startLogging(logfile=False):
    """
    Start logging with a custom formatter and optional file handler.

    This function initializes logging with a custom formatter for console output.
    If a logfile path is provided, it also sets up a file handler to log messages
    to the specified file. The logger captures memory allocation snapshots using
    `tracemalloc` for debugging purposes.

    Parameters:
    - logfile (str, optional): Path to a log file. If provided, logs will be written to this file.

    Returns:
    - logging.Logger: Configured logger object for logging.
    """
    tracemalloc.start()
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.DEBUG)
    sth = logging.StreamHandler()
    sth.setLevel(logging.INFO)
    sth.setFormatter(CustomFormatter())
    logger.addHandler(sth)
    if logfile:
        # remove if exists else will just keep appending
        if os.path.isfile(logfile):
            os.remove(logfile)
        fhnd = logging.FileHandler(logfile)
        fhnd.setLevel(logging.DEBUG)
        fhnd.setFormatter(FileFormatter())
        logger.addHandler(fhnd)
    return logger


def system_info(log):
    """
    Log system information including Python version and versions of funannotate2, gfftk, and buscolite.

    This function logs the current Python version along with the versions of the funannotate2,
    gfftk, and buscolite packages. It uses the provided logging function to output the information.

    Parameters:
    - log (function): A logging function to output the system information.

    Returns:
    - None
    """
    log(
        "Python v{}; helixerlite v{}, Helixer v0.3.2; HelixerPost v0.1.0, tensorflow v{}, h5py v{}, pyfastx v{}, gfftk v{}, numpy v{}".format(
            platform.python_version(),
            __version__,
            tf.__version__,
            h5py.__version__,
            pyfastx.__version__,
            gfftk.__version__,
            numpy.__version__,
        )
    )


def finishLogging(log, module):
    """
    Stop memory tracing, calculate peak memory usage, and log the module's completion.

    This function stops the memory tracing initiated by `tracemalloc`, calculates the peak
    memory usage during the module's execution, and logs this information using the provided
    logging function.

    Parameters:
    - log (function): A function used for logging messages.
    - module (str): The name of the module that has finished execution.

    Returns:
    - None
    """
    current, peak = tracemalloc.get_traced_memory()
    tracemalloc.stop()
    log(
        "{} module finished: peak memory usage={}".format(
            module, human_readable_size(peak)
        )
    )


def log_dependencies(script=False):
    return 0



import os
# silence tensorflow warnings
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
import warnings

with warnings.catch_warnings():
    warnings.filterwarnings("ignore")
    import tensorflow as tf
    from tensorflow.keras.models import Model
    from tensorflow.keras.layers import (
        Conv1D,
        LSTM,
        Dense,
        Bidirectional,
        Dropout,
        Reshape,
        Activation,
        Input,
        BatchNormalization,
    )
from .helixer_model import HelixerModel, HelixerSequence


class HybridSequence(HelixerSequence):
    def __init__(self, model, h5_files, mode, batch_size, shuffle):
        super().__init__(model, h5_files, mode, batch_size, shuffle)

    def __getitem__(self, idx):
        X, y, sw, transitions, phases, _, coverage_scores = self._generic_get_item(idx)

        if self.only_predictions:
            return X
        else:
            return X, y, sw


class HybridModel(HelixerModel):
    def __init__(self, cli_args=None):
        super().__init__(cli_args=cli_args)
        self.parser.add_argument("--cnn-layers", type=int, default=1)
        self.parser.add_argument("--lstm-layers", type=int, default=1)
        self.parser.add_argument("--units", type=int, default=32)
        self.parser.add_argument("--filter-depth", type=int, default=32)
        self.parser.add_argument("--kernel-size", type=int, default=26)
        self.parser.add_argument("--pool-size", type=int, default=9)
        self.parser.add_argument("--dropout1", type=float, default=0.0)
        self.parser.add_argument("--dropout2", type=float, default=0.0)
        self.parse_args()

    @staticmethod
    def sequence_cls():
        return HybridSequence

    def model(self):
        values_per_bp = 4
        if self.input_coverage:
            values_per_bp += self.coverage_count * 2

            raw_input = Input(
                shape=(None, values_per_bp),
                dtype=self.float_precision,
                name="raw_input",
            )
            main_input, coverage_input = tf.split(
                raw_input, [4, 2 * self.coverage_count], axis=-1
            )
            model_input = raw_input
        else:
            main_input = Input(
                shape=(None, values_per_bp),
                dtype=self.float_precision,
                name="main_input",
            )
            model_input = main_input
            coverage_input = None

        x = Conv1D(
            filters=self.filter_depth,
            kernel_size=self.kernel_size,
            padding="same",
            activation="relu",
        )(main_input)

        # if there are additional CNN layers
        for _ in range(self.cnn_layers - 1):
            x = BatchNormalization()(x)
            x = Conv1D(
                filters=self.filter_depth,
                kernel_size=self.kernel_size,
                padding="same",
                activation="relu",
            )(x)

        if self.pool_size > 1:
            x = Reshape((-1, self.pool_size * self.filter_depth))(x)
            # x = MaxPooling1D(pool_size=self.pool_size, padding='same')(x)

        if self.dropout1 > 0.0:
            x = Dropout(self.dropout1)(x)

        x = Bidirectional(LSTM(self.units, return_sequences=True))(x)
        for _ in range(self.lstm_layers - 1):
            x = Bidirectional(LSTM(self.units, return_sequences=True))(x)

        # do not use recurrent dropout, but dropout on the output of the LSTM stack
        if self.dropout2 > 0.0:
            x = Dropout(self.dropout2)(x)

        outputs = self.model_hat((x, coverage_input))

        model = Model(inputs=model_input, outputs=outputs)
        return model

    def model_hat(self, penultimate_layers):
        x, coverage_input = penultimate_layers
        # maybe concatenate coverage and add one extra dense at this point
        if self.input_coverage:
            coverage_input = Reshape((-1, self.pool_size * self.coverage_count * 2))(
                coverage_input
            )
            x = tf.concat([x, coverage_input], axis=-1)
            if self.post_coverage_hidden_layer:
                x = Dense(self.units // 2)(x)

        if self.predict_phase:
            x = Dense(self.pool_size * 4 * 2)(x)  # predict twice a many floats
            x_genic, x_phase = tf.split(x, 2, axis=-1)

            x_genic = Reshape((-1, self.pool_size, 4), name="reshape_hat")(x_genic)
            x_genic = Activation("softmax", name="genic")(x_genic)

            x_phase = Reshape((-1, self.pool_size, 4), name="reshape_hat1")(x_phase)
            x_phase = Activation("softmax", name="phase")(x_phase)

            outputs = [x_genic, x_phase]
        else:
            x = Dense(self.pool_size * 4)(x)
            x = Reshape((-1, self.pool_size, 4), name="reshape_hat")(x)
            x = Activation("softmax", name="main")(x)
            outputs = [x]

        return outputs

    def compile_model(self, model):
        if self.predict_phase:
            losses = ["categorical_crossentropy", "categorical_crossentropy"]
            loss_weights = [0.8, 0.2]
        else:
            losses = ["categorical_crossentropy"]
            loss_weights = [1.0]

        model.compile(
            optimizer=self.optimizer,
            loss=losses,
            loss_weights=loss_weights,
            sample_weight_mode="temporal",
        )



#!/usr/bin/env python3

import sys
import os
import argparse
import uuid
import json
from gfftk.gff import gff2dict, dict2gff3
from gfftk.stats import annotation_stats
from .__init__ import __version__
from .help_formatter import MyParser, MyHelpFormatter
from .log import startLogging, system_info, finishLogging
from .hdf5 import HelixerFastaToH5Controller
from .hybrid_model import HybridModel
from .utilities import preds2gff3, download


def main():
    args = parse_args(sys.argv[1:])
    logger = startLogging(logfile=f"{args.out.rsplit('.', 1)[0]}.log")
    system_info(logger.info)
    config = {
        "fungi": {
            "url": "https://zenodo.org/records/10836346/files/fungi_v0.3_a_0100.h5?download=1",
            "name": "fungi_v0.3_a_0100.h5",
            "length": 21384,
            "offset": 10692,
            "core": 16038,
        },
        "land_plants": {
            "url": "https://zenodo.org/records/10836346/files/land_plant_v0.3_a_0080.h5?download=1",
            "name": "and_plant_v0.3_a_0080.h5",
            "length": 54152,
            "offset": 32076,
            "core": 48114,
        },
        "vertebrate": {
            "url": "https://zenodo.org/records/10836346/files/vertebrate_v0.3_a_0200.h5?download=1",
            "name": "vertebrate_v0.3_a_0200.h5",
            "length": 213840,
            "offset": 106920,
            "core": 160380,
        },
        "invertebrate": {
            "url": "https://zenodo.org/records/10836346/files/invertebrate_v0.3_a_0300.h5?download=1",
            "name": "invertebrate_v0.3_a_0300.h5",
            "length": 213840,
            "offset": 106920,
            "core": 160380,
        },
    }
    lineage_config = config.get(args.lineage)
    # check if model exists
    global_model = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        "model_files",
        lineage_config["name"],
    )
    if not os.path.isfile(global_model):
        if os.access(os.path.dirname(global_model), os.W_OK):
            model_path = global_model
        else:  # cannot write there, so put in working directory
            logger.info(
                f"User does not appear to have write access to {os.path.dirname(global_model)}, will download model to current directory."
            )
            model_path = lineage_config["name"]
    else:
        logger.info(f"Helixer model already exists: {global_model}")
        model_path = global_model

    if not os.path.isfile(model_path):
        logger.info(f"Downloading Helixer model: {model_path}")
        download(lineage_config["url"], model_path)
    # convert genome to hd5
    slug = str(uuid.uuid4())[:8]
    logger.info(
        f"Converting genome assembly to HDF5 format. Using settings for {args.lineage}: --subsequence-length {lineage_config['length']}"
    )
    fasta2hdf5(
        args.fasta, f"{slug}.h5", species=slug, subseqlen=lineage_config["length"]
    )
    # train with hybrid model
    # build cli options
    # --load-model-path fungi_v0.3_a_0100.h5 --test-data 5f616493-e12d-4090-aace-df674baaafe3.h5 --overlap --val-test-batch-size 32
    logger.info(f"Running helixer HybridModel using model={lineage_config['name']}")
    logger.info(
        f'Using settings for {args.lineage}: --overlap-offset {lineage_config["offset"]} --core-length {lineage_config["core"]}'
    )
    model = HybridModel(
        [
            "--load-model-path",
            model_path,
            "--test-data",
            f"{slug}.h5",
            "--overlap-offset",
            str(lineage_config["offset"]),
            "--core-length",
            str(lineage_config["core"]),
            "--overlap",
            "--val-test-batch-size",
            "32",
            "--prediction-output-path",
            f"{slug}.predictions.h5",
            "--cpus",
            str(args.cpus),
        ]
    )
    model.run()

    # convert to GFF3
    logger.info("Converting predictions.h5 to GFF3 with helixerpost")
    preds2gff3(
        f"{slug}.h5",
        f"{slug}.predictions.h5",
        f"{slug}.gff3",
        peak_threshold=float(args.peak),
        min_coding_length=args.minprotlen,
    )
    # clean/reformat gff
    logger.info("Cleaning GFF3 output with GFFtk")
    Genes = gff2dict(f"{slug}.gff3", args.fasta)
    stats = annotation_stats(Genes)
    dict2gff3(Genes, output=args.out)
    logger.info(f"Annotation stats:\n{json.dumps(stats, indent=2)}")

    # clean up
    for f in [f"{slug}.h5", f"{slug}.predictions.h5", f"{slug}.gff3"]:
        if os.path.isfile(f):
            os.remove(f)
    # finish
    finishLogging(logger.info, vars(sys.modules[__name__])["__name__"])


def parse_args(args):
    description = "helixerlite: simplified helixer de novo genome annotation"
    parser = MyParser(
        description=description, formatter_class=MyHelpFormatter, add_help=False
    )
    required = parser.add_argument_group("Required arguments")
    optional = parser.add_argument_group("Optional arguments")

    required.add_argument(
        "-f",
        "--fasta",
        required=True,
        metavar="",
        help="Input sequence file in FASTA format",
    )
    required.add_argument(
        "-o",
        "--out",
        required=True,
        metavar="",
        help="GFF3 output",
    )
    required.add_argument(
        "-l",
        "--lineage",
        required=True,
        metavar="",
        choices=["fungi", "land_plant", "vertebrate", "invertebrate"],
        help="Specify which model to use for prediction [fungi,land_plant,vertebrate,invertebrate]",
    )
    optional.add_argument(
        "-c",
        "--cpus",
        required=False,
        type=int,
        default=1,
        metavar="",
        help="Specify the number (N=integer) of threads/cores to use.",
    )
    optional.add_argument(
        "-p",
        "--peak-threshold",
        dest="peak",
        required=False,
        default=0.8,
        metavar="",
        help="Minimum peak genic score to call a gene model. [< 1]",
    )
    optional.add_argument(
        "-m",
        "--min-protein-length",
        dest="minprotlen",
        required=False,
        default=60,
        metavar="",
        help="Minimum length protein coding gene [amino acids]",
    )
    help_args = parser.add_argument_group("Help")
    help_args.add_argument(
        "-h",
        "--help",
        action="help",
        default=argparse.SUPPRESS,
        help="Show this help message and exit",
    )
    help_args.add_argument(
        "--version",
        action="version",
        version="{} v{}".format(
            os.path.basename(os.path.dirname(os.path.realpath(__file__))), __version__
        ),
        help="show program's version number and exit",
    )

    # If no arguments were used, print the base-level help which lists possible commands.
    if len(args) == 0:
        parser.print_help(file=sys.stderr)
        raise SystemExit(1)

    return parser.parse_args(args)


def fasta2hdf5(fasta, hdout, subseqlen=21384, species=str(uuid.uuid4())):
    controller = HelixerFastaToH5Controller(fasta, hdout)
    controller.export_fasta_to_h5(
        chunk_size=subseqlen,
        compression="gzip",
        multiprocess=True,
        species=species,
        write_by=20_000_000,
    )


if __name__ == "__main__":
    main()



"""convert cleaned-db schema to numeric values describing gene structure"""

import time

import numpy as np
import math
import multiprocessing
from abc import ABC, abstractmethod


AMBIGUITY_DECODE = {
    "C": [1.0, 0.0, 0.0, 0.0],
    "A": [0.0, 1.0, 0.0, 0.0],
    "T": [0.0, 0.0, 1.0, 0.0],
    "G": [0.0, 0.0, 0.0, 1.0],
    "Y": [0.5, 0.0, 0.5, 0.0],
    "R": [0.0, 0.5, 0.0, 0.5],
    "W": [0.0, 0.5, 0.5, 0.0],
    "S": [0.5, 0.0, 0.0, 0.5],
    "K": [0.0, 0.0, 0.5, 0.5],
    "M": [0.5, 0.5, 0.0, 0.0],
    "D": [0.0, 0.33, 0.33, 0.33],
    "V": [0.33, 0.33, 0.0, 0.33],
    "H": [0.33, 0.33, 0.33, 0.0],
    "B": [0.33, 0.0, 0.33, 0.33],
    "N": [0.25, 0.25, 0.25, 0.25],
}


class Stepper(object):
    def __init__(self, end, by):
        self.at = 0
        self.end = end
        self.by = by

    def step(self):
        prev = self.at
        if prev + self.by < self.end:
            new = prev + self.by
        else:
            new = self.end
        self.at = new
        return prev, new

    def step_to_end(self):
        while self.at < self.end:
            yield self.step()


class Numerifier(ABC):
    def __init__(self, n_cols, coord, max_len, dtype=np.float32, start=0, end=None):
        if end is None:
            end = coord.length
        assert isinstance(n_cols, int)
        self.n_cols = n_cols
        self.coord = coord
        self.max_len = max_len
        self.dtype = dtype
        self.matrix = None
        self.start = start
        self.end = end
        self.length = self.end - self.start
        # set paired steps
        partitioner = Stepper(end=self.end - self.start, by=self.max_len)
        self.paired_steps = list(partitioner.step_to_end())
        super().__init__()

    @abstractmethod
    def coord_to_matrices(self):
        """Method to be called from outside. Numerifies both strands."""
        pass

    def _slice_matrices(self, is_plus_strand, *argv):
        """Slices (potentially) multiple matrices in the same way according to self.paired_steps"""
        assert len(argv) > 0, "Need a matrix to slice"
        all_slices = [[] for _ in range(len(argv))]
        # reverse steps on minus strand
        steps = self.paired_steps if is_plus_strand else self.paired_steps[::-1]
        for prev, current in steps:
            for matrix, slices in zip(argv, all_slices):
                data_slice = matrix[prev:current]
                if not is_plus_strand:
                    # invert directions
                    data_slice = np.flip(data_slice, axis=0)
                slices.append(data_slice)
        return all_slices

    def _zero_matrix(self):
        self.matrix = np.zeros(
            (
                self.length,
                self.n_cols,
            ),
            self.dtype,
        )


def seq_numerify(seq_part):
    as_list = [AMBIGUITY_DECODE[c.upper()] for c in seq_part]
    return np.array(as_list, np.float16)


class SequenceNumerifier(Numerifier):
    def __init__(self, coord, max_len, start=0, end=None, use_multiprocess=True):
        self.use_multiprocess = use_multiprocess
        super().__init__(
            n_cols=4,
            coord=coord,
            max_len=max_len,
            dtype=np.float16,
            start=start,
            end=end,
        )

    def coord_to_matrices(self):
        """Does not alter the error mask unlike in AnnotationNumerifier"""

        # plus strand, actual numerification of the sequence
        start_time = time.time()
        seq = self.coord.sequence[self.start : self.end]
        seq_len = len(seq)  # can be slow
        if seq_len < int(1e6) or not self.use_multiprocess:
            # numerify short sequences sequentially
            as_list = [AMBIGUITY_DECODE[c.upper()] for c in seq]
            self.matrix = np.array(as_list, self.dtype)

        else:
            # numerify longer sequences in parallel
            # use one less cpu than possible, with minimum 500K chars per process
            n_processes = min(multiprocessing.cpu_count(), seq_len // int(5e5)) - 1
            with multiprocessing.Pool(n_processes) as p:
                max_seq_part_len = int(np.ceil(seq_len / n_processes))
                seq_parts = [
                    seq[offset : offset + max_seq_part_len]
                    for offset in range(0, seq_len, max_seq_part_len)
                ]
                numerified_parts = p.map(seq_numerify, seq_parts)
            assert seq_len == sum([len(p) for p in numerified_parts])
            self.matrix = np.concatenate(numerified_parts)

        # very important to copy here
        data_plus = self._slice_matrices(True, np.copy(self.matrix))[0]

        # minus strand
        self.matrix = np.flip(self.matrix, axis=1)  # complementary base
        # slice matrix will reverse direction
        data_minus = self._slice_matrices(False, self.matrix)[0]

        # put everything together
        data = {"plus": data_plus, "minus": data_minus}
        return data


class MatAndInfo:
    """organizes data and meta info for post-processing and saving a matrix"""

    def __init__(self, key, matrix, dtype):
        self.key = key
        self.matrix = matrix
        self.dtype = dtype

    def __repr__(self):
        return "key: {}, matrix shape: {}, matrix dtype {}: target dtype {}".format(
            self.key, self.matrix.shape, self.matrix.dtype, self.dtype
        )


class CoordNumerifier(object):
    """Combines the different Numerifiers which need to operate on the same Coordinate
    to ensure consistent parameters. Selects all Features of the given Coordinate.
    """

    @staticmethod
    def pad(d, chunk_size):
        n_seqs = len(d)
        # insert all the sequences so that 0-padding is added if needed
        shape = tuple([n_seqs, chunk_size] + list(d[0].shape[1:]))
        padded_d = np.zeros(shape, dtype=d[0].dtype)
        for j in range(n_seqs):
            padded_d[j, : len(d[j])] = d[j]
        return padded_d

    @staticmethod
    def start_ends(numerifier, strand):
        assert isinstance(numerifier, Numerifier)
        if strand == "plus":
            start_ends = numerifier.paired_steps
        else:
            # flip the start ends back for - strand
            start_ends = [(x[1], x[0]) for x in numerifier.paired_steps[::-1]]
        start_ends = np.array(start_ends, dtype=np.int64)
        return start_ends

    @staticmethod
    def seq_matinfos(coord, genome, start_ends, length):
        res = [
            MatAndInfo("species", np.array([genome.encode("ASCII")] * length), "S25"),
            MatAndInfo(
                "seqids", np.array([coord.seqid.encode("ASCII")] * length), "S50"
            ),
            MatAndInfo("start_ends", start_ends, "int64"),
        ]
        return res

    @staticmethod
    def numerify_only_fasta(
        coord, max_len, genome, one_hot=True, use_multiprocess=False, write_by=20000000
    ):
        """export the FASTA sequence only"""
        # passing empty features causes SplitFinder to consider noting more than splitting
        # to max length of write_by and end of sequence handling.
        split_finder = SplitFinder(
            features=(),
            write_by=write_by,
            coord_length=coord.length,
            chunk_size=max_len,
        )
        for _, bp_coord, h5_coord in split_finder.feature_n_coord_gen():
            start, end = bp_coord
            seq_numerifier = SequenceNumerifier(
                coord=coord,
                max_len=max_len,
                start=start,
                end=end,
                use_multiprocess=use_multiprocess,
            )
            xb = seq_numerifier.coord_to_matrices()
            for strand in ["plus", "minus"]:
                x = CoordNumerifier.pad(xb[strand], max_len)
                start_ends = CoordNumerifier.start_ends(seq_numerifier, strand)
                start_ends += start
                out = [MatAndInfo("X", x, "float16")]
                out.extend(
                    CoordNumerifier.seq_matinfos(coord, genome, start_ends, len(x))
                )
                yield tuple(out), h5_coord[strand]

    @staticmethod
    def numerify(
        coord,
        coord_features,
        max_len,
        one_hot=True,
        mode=("X", "y", "anno_meta", "transitions"),
        write_by=5000000,
        use_multiprocess=True,
    ):
        assert isinstance(max_len, int) and max_len > 0, "what is {} of type {}".format(
            max_len, type(max_len)
        )
        coord_features = sorted(
            coord_features, key=lambda f: min(f.start, f.end)
        )  # sort by ~ +strand start
        split_finder = SplitFinder(
            features=coord_features,
            write_by=write_by,
            coord_length=coord.length,
            chunk_size=max_len,
        )
        for f_set, bp_coord, h5_coord in split_finder.feature_n_coord_gen():
            for strand_res in CoordNumerifier._numerify_super_write_chunk(
                f_set,
                bp_coord,
                h5_coord,
                coord,
                max_len,
                one_hot,
                coord_features,
                mode,
                use_multiprocess,
            ):
                yield strand_res


# todo, consider moving to separate splitting file or exporter...?
class SplitFinder:
    def __init__(self, features, write_by, coord_length, chunk_size):
        if write_by % chunk_size:
            old_write_by = write_by
            write_by = chunk_size * (write_by // chunk_size)
        self.features = features
        self.write_by = write_by  # target writing this many bp to the h5 file at once
        self.coord_length = coord_length
        self.chunk_size = chunk_size
        self.splits = tuple(self._find_splits())
        self.relative_h5_coords = tuple(self._get_rel_h5_coords_for_splits())

    @property
    def coords(self):
        starts = [0] + list(self.splits)[:-1]
        return tuple(zip(starts, self.splits))

    def feature_n_coord_gen(self):
        """generator for feature subset, bp coordinates, and h5 coordinates (for +/-)"""
        return zip(self.split_features(), self.coords, self.relative_h5_coords)

    def split_features(self):
        """get all features from start of list that aren't passed _to_"""
        i = 0
        in_split = []
        in_next_split = []
        for end in self.splits:
            feature = self._ith_feature_or_none(i)
            while self._feature_not_past(feature, end):
                in_split.append(feature)
                if self._feature_ends_after(
                    feature, end
                ):  # basically this is 'if overlaps end' in context
                    in_next_split.append(feature)
                i += 1
                feature = self._ith_feature_or_none(i)
            yield in_split
            in_split = in_next_split
            in_next_split = []

    def _ith_feature_or_none(self, i):
        try:
            return self.features[i]
        except IndexError:
            return None

    def _get_rel_h5_coords_for_splits(self):
        """calculates where to write the +/- strand super-chunk splits in the h5 file"""
        # calculate the positive strand first
        postive_h5_ends = []
        postive_h5_starts = [0]
        for end in self.splits:
            p_h5 = end // self.chunk_size
            if end % self.chunk_size:  # end of seq, will be padded to full chunk size
                p_h5 += 1
            else:
                postive_h5_starts.append(p_h5)
            postive_h5_ends.append(p_h5)

        postive_h5s = list(zip(postive_h5_starts, postive_h5_ends))
        # calculate the negative strand as positive strand, but backwards from end
        h5_end = postive_h5_ends[-1] * 2
        negative_h5s = [(h5_end - x[1], h5_end - x[0]) for x in postive_h5s]
        return ({"plus": x[0], "minus": x[1]} for x in zip(postive_h5s, negative_h5s))

    @staticmethod
    def _feature_not_past(feature, to):
        """whether feature is before or overlaps end by any measure"""
        if feature is None:
            return False  # force break of while loop when out of features
        if feature.is_plus_strand:
            return feature.start < to
        else:
            # although the end is exclusive and this gets one position where the feature itself has no overlap;
            # for transitions themselves we potentially want to make sure the feature.end is included
            return feature.end < to

    @staticmethod
    def _feature_ends_after(feature, to):
        """combines with _feature_not_past to define overlaps of trailing edge 'to'"""
        # overlapping features will be saved and numerified with the next write_by split as well
        if feature.is_plus_strand:
            # include as soon as end is after, despite exclusive end, to include the transition just in case
            return feature.end >= to
        else:
            return feature.start >= to

    def _find_splits(self):
        """yields splits of ~write_by size that can be safely split at"""
        tr_mask = self._transition_and_split_cds_mask()
        for i in range(self.write_by, self.coord_length, self.write_by):
            if i not in tr_mask:
                yield i
            else:
                for i_fudge in range(
                    i, min(i + self.write_by, self.coord_length), self.chunk_size
                ):
                    if i_fudge not in tr_mask:
                        yield i_fudge
                        break
        yield self.coord_length

    def _transition_and_split_cds_mask(self):
        """mark all possible splits where there is a transition, so splitting there would change the numerify results"""
        tr_mask = set()
        for feature in self.features:
            # avoid splitting exactly at transitions, as transitions are detected by
            # state change (of binary encoding) and you can't detect state-change
            # if you split at it
            f_start, f_end = self._plus_strand_transitions(feature)
            for tr in (f_start, f_end):
                if not tr % self.chunk_size:
                    tr_mask.add(tr)
            # also avoid splitting in the middle of a CDS, this is necessary to
            # simplify the encoding of phase / avoid painful edge cases
            # this should add all chunk ends with a CDS to the mask
            cs = self.chunk_size
            if feature.type.value == "geenuff_cds":
                round_start = math.ceil(f_start / cs)
                round_end = math.ceil(f_end / cs)
                for chunk_end in range(round_start * cs, round_end * cs, cs):
                    tr_mask.add(chunk_end)

        return tr_mask

    @staticmethod
    def _plus_strand_transitions(feature):
        if feature.is_plus_strand:
            return feature.start, feature.end
        else:
            return feature.end - 1, feature.start - 1



import sys

import numpy as np
import os
import csv
from collections import defaultdict
from scipy.sparse import coo_matrix


class ConfusionMatrix:
    def __init__(self, col_names=None, skip_uncertainty=True):
        if col_names is None:
            col_names = ["ig", "utr", "exon", "intron"]
        self.col_names = {i: name for i, name in enumerate(col_names)}
        self.skip_uncertainty = skip_uncertainty
        self.n_classes = len(self.col_names)
        self.cm = np.zeros((self.n_classes, self.n_classes), dtype=np.uint64)
        if not self.skip_uncertainty:
            self.uncertainties = {
                col_name: list() for col_name in self.col_names.values()
            }
            self.max_uncertainty = -sum(
                [e * np.log2(e) for e in [1 / self.n_classes] * self.n_classes]
            )

    @staticmethod
    def _argmax_y(arr):
        arr = np.argmax(arr, axis=-1).ravel().astype(np.int8)
        return arr

    @staticmethod
    def _remove_masked_bases(y_true, y_pred, sw):
        """Remove bases marked as errors, should also remove zero padding"""
        sw = sw.astype(bool)
        y_true = y_true[sw]
        y_pred = y_pred[sw]
        return y_true, y_pred

    def _add_to_cm(self, y_true, y_pred):
        """Put in extra function to be testable"""
        # add to confusion matrix as long as _some_ bases were not masked
        if y_pred.size > 0:
            y_pred = ConfusionMatrix._argmax_y(y_pred)
            y_true = ConfusionMatrix._argmax_y(y_true)
            # taken from here, without the boiler plate:
            # https://github.com/scikit-learn/scikit-learn/blob/
            # 42aff4e2edd8e8887478f6ff1628f27de97be6a3/sklearn/metrics/_classification.py#L342
            cm_batch = coo_matrix(
                (np.ones(y_true.shape[0], dtype=np.int8), (y_true, y_pred)),
                shape=(self.n_classes, self.n_classes),
                dtype=np.uint32,
            ).toarray()
            self.cm += cm_batch

    def _add_to_uncertainty(self, y_true, y_pred):
        y_pred = y_pred.reshape((-1, y_pred.shape[-1]))
        y_true = ConfusionMatrix._argmax_y(y_true)
        # entropy calculation
        y_pred_log2 = np.log2(y_pred)
        y_pred_H = -1 * np.sum(y_pred * y_pred_log2, axis=-1)
        # average entropy for all the bases in one class according to the labels
        for i, name in self.col_names.items():
            class_mask = y_true == i
            if np.any(class_mask):
                avg_entropy = np.nanmean(y_pred_H[class_mask])
                avg_entropy /= (
                    self.max_uncertainty
                )  # normalize by maximum for comparability
                self.uncertainties[name].append(avg_entropy)

    def count_and_calculate_one_batch(self, y_true, y_pred, sw):
        y_true, y_pred = ConfusionMatrix._remove_masked_bases(y_true, y_pred, sw)
        if not self.skip_uncertainty:
            # important to copy so _add_to_cm() works
            self._add_to_uncertainty(np.copy(y_true), np.copy(y_pred))
        self._add_to_cm(y_true, y_pred)

    def _get_normalized_cm(self):
        """Put in extra function to be testable"""
        class_sums = np.sum(self.cm, axis=1)
        normalized_cm = (
            self.cm / class_sums[:, None]
        )  # expand by one dim so broadcast work properly
        return normalized_cm

    @staticmethod
    def _precision_recall_f1(tp, fp, fn):
        if (tp + fp) > 0:
            precision = tp / (tp + fp)
        else:
            precision = 0.0  # avoid an error due to division by 0
        if (tp + fn) > 0:
            recall = tp / (tp + fn)
        else:
            recall = 0.0
        if (precision + recall) > 0:
            f1 = 2 * precision * recall / (precision + recall)
        else:
            f1 = 0.0
        return precision, recall, f1

    def _total_accuracy(self):
        return np.trace(self.cm) / np.sum(self.cm)

    @staticmethod
    def _add_to_scores(d):
        """adds precision, recall, and f1 to input dictionary, d"""
        metrics = ConfusionMatrix._precision_recall_f1(d["TP"], d["FP"], d["FN"])
        d["precision"], d["recall"], d["f1"] = metrics

    def _get_scores(self):
        scores = defaultdict(dict)
        # single column metrics
        for col in range(self.n_classes):
            name = self.col_names[col]
            d = scores[name]
            not_col = np.arange(self.n_classes) != col
            # reference annotation is in dim 0, prediction in dim 1
            d["TP"] = self.cm[col, col]
            d["FP"] = np.sum(self.cm[not_col, col])
            d["FN"] = np.sum(self.cm[col, not_col])
            # add average uncertanties
            if not self.skip_uncertainty:
                d["H"] = np.mean(self.uncertainties[name])

            ConfusionMatrix._add_to_scores(d)
        return scores

    def prep_tables(self, scores):
        out = []
        names = list(self.col_names.values())

        # confusion matrix
        cm = [[""] + [x + "_pred" for x in names]]
        for i, row in enumerate(self.cm.astype(int).tolist()):
            cm.append([names[i] + "_ref"] + row)
        out.append((cm, "confusion_matrix"))

        # normalized
        normalized_cm = [cm[0]]
        for i, row in enumerate(self._get_normalized_cm().tolist()):
            normalized_cm.append(
                [names[i] + "_ref"] + [round(x, ndigits=4) for x in row]
            )
        out.append((normalized_cm, "normalized_confusion_matrix"))

        # F1
        table = [["", "norm. H", "Precision", "Recall", "F1-Score"]]
        for i, (name, values) in enumerate(scores.items()):
            # check if there is an entropy value coming (only for single type classes)
            if not self.skip_uncertainty and i < len(names):
                metrics = []
            else:
                metrics = [""]
            metrics += [
                "{:.4f}".format(s) for s in list(values.values())[3:]
            ]  # [3:] to skip TP, FP, FN
            table.append([name] + metrics)
            if i == (len(names) - 1) and len(scores) > len(names):
                table.append([""] * 4)
        out.append((table, "F1_summary"))

        return out

    def export_to_csvs(self, pathout):
        if pathout is not None:
            if not os.path.exists(pathout):
                os.makedirs(pathout)
            scores = self._get_scores()
            for table, table_name in self.prep_tables(scores):
                with open("{}/{}.csv".format(pathout, table_name), "w") as f:
                    writer = csv.writer(f)
                    for row in table:
                        writer.writerow(row)


class ConfusionMatrixGenic(ConfusionMatrix):
    """Extension of ConfusionMatrix that just adds the calculation of the composite scores"""

    def _get_scores(self):
        scores = super()._get_scores()

        # legacy cds score that works the same as the cds_f1 with the 3 column multi class encoding
        # essentiall merging the predictions as if error between exon and intron did not matter
        d = scores["legacy_cds"]
        cm = self.cm
        d["TP"] = cm[2, 2] + cm[3, 3] + cm[2, 3] + cm[3, 2]
        d["FP"] = cm[0, 2] + cm[0, 3] + cm[1, 2] + cm[1, 3]
        d["FN"] = cm[2, 0] + cm[3, 0] + cm[2, 1] + cm[3, 1]
        ConfusionMatrix._add_to_scores(d)

        # subgenic metric is essentially the same as the genic one
        # pretty redundant code to below, but done for minimizing the risk to mess up (for now)
        d = scores["sub_genic"]
        for base_metric in ["TP", "FP", "FN"]:
            d[base_metric] = sum([scores[m][base_metric] for m in ["exon", "intron"]])
        ConfusionMatrix._add_to_scores(d)

        # genic metrics are calculated by summing up TP, FP, FN, essentially calculating a weighted
        # sum for the individual metrics. TP of the intergenic class are not taken into account
        d = scores["genic"]
        for base_metric in ["TP", "FP", "FN"]:
            d[base_metric] = sum(
                [scores[m][base_metric] for m in ["utr", "exon", "intron"]]
            )
        ConfusionMatrix._add_to_scores(d)

        return scores


class ConfusionMatrixPhase(ConfusionMatrix):
    """Extension of ConfusionMatrix to differentiate phase shift from CDS vs not mistake"""

    def __init__(self, skip_uncertainty=True):
        super().__init__(
            col_names=["no_phase", "phase_0", "phase_1", "phase_2"],
            skip_uncertainty=skip_uncertainty,
        )

    def _get_scores(self):
        scores = super()._get_scores()
        # scores is a defaultdict, of dictionaries

        # any_phase
        # basically phase vs not phase
        # should be close to CDS score, but with a slightly lower recall
        # more false negatives are always expected in the 2x4 encoding, e.g. [0.35, 0.2, 0.2, 0.25]
        # which will end up being CDS w/ phase_2 in the end, but argmax here will have given no_phase
        d = scores["any_phase"]
        cm = self.cm
        # sum up any_phase predictions
        d["TP"] = np.sum(cm[1:, 1:])
        d["FP"] = np.sum(cm[0, 1:])
        d["FN"] = np.sum(cm[1:, 0])
        ConfusionMatrix._add_to_scores(d)

        # internal_phase
        # to detect shifted phase predictions, as opposed to phase not phase
        d = scores["internal_phase"]
        d["TP"] = cm[1, 1] + cm[2, 2] + cm[3, 3]
        # OK, as all are now 'positive classes of interest' this will give us the _accuracy_!
        # but that's fine because within phase there's class balance, and it keeps the implementation consistent...
        mask = np.logical_not(np.eye(3))
        n_mistakes = np.sum(cm[1:, 1:][mask])
        d["FP"] = n_mistakes
        d["FN"] = n_mistakes
        ConfusionMatrix._add_to_scores(d)

        # phase
        # this is a lot like CDS F1, but with frame shifts still being penalized (as well as qualifiers above)
        d = scores["phase"]
        for base_metric in ["TP", "FP", "FN"]:
            d[base_metric] = sum(
                [scores[m][base_metric] for m in ["phase_0", "phase_1", "phase_2"]]
            )
        ConfusionMatrix._add_to_scores(d)

        return scores


class Metrics:
    def __init__(self, generator, print_to_stdout=True, skip_uncertainty=True):
        np.set_printoptions(
            suppress=True
        )  # do not use scientific notation for the print out
        self.generator = generator
        self.print_to_stdout = print_to_stdout
        self.skip_uncertainty = skip_uncertainty
        self.cm_genic = ConfusionMatrixGenic(skip_uncertainty=self.skip_uncertainty)
        self.cm_phase = ConfusionMatrix(
            ["no_phase", "phase_0", "phase_1", "phase_2"],
            skip_uncertainty=self.skip_uncertainty,
        )

    def _overlap_all_data(self, batch_idx, y_true, y_pred, sw):
        assert len(y_pred.shape) == 4, (
            "this reshape assumes shape is "
            "(batch_size, chunk_size // pool, pool, label dim)"
            "and apparently it is time to fix that, shape is {}".format(y_pred.shape)
        )
        bs, cspool, pool, ydim = y_pred.shape
        y_pred = y_pred.reshape([bs, cspool * pool, ydim])
        y_pred = self.generator.ol_helper.overlap_predictions(batch_idx, y_pred)
        y_pred = y_pred.reshape([-1, cspool, pool, ydim])
        # edge handle sw & y_true (as done with y_pred and to restore 1:1 input output
        sw = self.generator.ol_helper.subset_input(batch_idx, sw)
        y_true = self.generator.ol_helper.subset_input(batch_idx, y_true)
        return y_true, y_pred, sw

    def calculate_metrics(self, model):
        for batch_idx in range(len(self.generator)):
            print(batch_idx, "/", len(self.generator) - 1, end="\r")

            inputs = self.generator[batch_idx]
            if len(inputs) == 2 and type(inputs[0]) is list:
                mode = "dialated_conv"
                (X, sw), y_true = inputs
                y_pred = model.predict_on_batch([X, sw])
            elif len(inputs) == 3:
                if type(inputs[0]) is list:
                    mode = "correction"
                    (X, pred), y_true, sw = inputs
                    y_pred = model.predict_on_batch([X, pred])
                elif type(inputs[1]) is list:
                    mode = "phase"
                    X, (y_true, y_true_phase), sw = inputs
                    y_pred, y_pred_phase = model.predict_on_batch(X)
                else:
                    mode = "regular"
                    X, y_true, sw = inputs
                    y_pred = model.predict_on_batch(X)
            else:
                print(f"Unknown inputs from keras sequence: {inputs}", file=sys.stderr)
                exit()

            data = {"genic_base_wise": [self.cm_genic, (y_true, y_pred)]}
            if mode == "phase":
                data["phase_base_wise"] = [self.cm_phase, (y_true_phase, y_pred_phase)]

            all_scores = {}
            for _, (cm, (y_true, y_pred)) in data.items():
                sw_copy = sw.copy()
                if self.generator.overlap:
                    y_true, y_pred, sw_copy = self._overlap_all_data(
                        batch_idx, y_true, y_pred, sw_copy
                    )
                cm.count_and_calculate_one_batch(y_true, y_pred, sw_copy)

        # data contains cms + metric
        for metric_name, (cm, (_, _)) in data.items():
            scores = cm._get_scores()
            all_scores[metric_name] = scores
        return all_scores
